###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import functools
import re
import sys
from collections.abc import Callable, Collection
from dataclasses import dataclass
from itertools import pairwise
from typing import TYPE_CHECKING, Generic, TypeVar

import hyperscan


if TYPE_CHECKING:
    from .job_analyzer import Job


T = TypeVar("T")


@dataclass
class LogFileMatches:
    init: list[LogFileMatch]
    by_lfn: dict[str, list[LogFileMatch]]
    finalize: list[LogFileMatch]

    def __init__(self, lfns: Collection[str]) -> None:
        self.init = []
        self.by_lfn = {lfn: [] for lfn in lfns}
        self.finalize = []

    def apply_sorting(self):
        self.init = sorted(self.init)
        self.by_lfn = {k: sorted(v) for k, v in self.by_lfn.items()}
        self.finalize = sorted(self.finalize)


class classproperty(Generic[T]):
    def __init__(self, method: Callable[..., T]):
        self.method = method
        functools.update_wrapper(self, method)  # type: ignore

    def __get__(self, obj, cls=None) -> T:
        if cls is None:
            cls = type(obj)
        return self.method(cls)


class ReasonMixin:
    """Base class for mixins that customize failure reason behavior.

    Mixins can override specific methods to provide custom implementations
    for failure reason classes. Only the methods defined here can be overridden.
    """

    @property
    def match_string(self) -> str:
        """Return the string representation of the match.

        Override this to customize how the matched portion of the log is extracted
        and formatted. The default implementation returns the matched bytes decoded
        as a string.
        """
        raise NotImplementedError("Subclasses must implement match_string")


class LogFileMatch:
    pattern_flags: int = hyperscan.HS_FLAG_SOM_LEFTMOST
    explanation: str | None = None
    _db: hyperscan.Database | None = None

    _reasons = []
    id_to_pattern: dict[int, str] = {}
    id_to_reason: dict[int, type[LogFileMatch]] = {}

    def __init__(self, job: Job, start: int, end: int) -> None:
        """Initialize a specific occurrence of this failure."""
        self._job = job
        self._start = start
        self._end = end

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(job, {self._start}, {self._end})"

    def __hash__(self) -> int:
        return hash((self.__class__, self._job, self.match_string))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self._job == other._job and self.match_string == other.match_string

    @property
    def match_string(self) -> str:
        return self._job.files.steps[-1].application_log[self._start : self._end].decode(errors="backslashreplace")

    @classproperty
    def reason_to_id(cls) -> dict[type[LogFileMatch], int]:  # pylint: disable=no-self-argument
        if not hasattr(cls, "_reason_to_id"):
            # Trigger the initialization of the database
            cls.db
        return cls._reason_to_id

    @classproperty
    def name_to_id(cls) -> dict[str, int]:  # pylint: disable=no-self-argument
        if not hasattr(cls, "_name_to_id"):
            # Trigger the initialization of the database
            cls.db
        return cls._name_to_id

    @classproperty
    def name_to_reason(cls) -> dict[str, type[LogFileMatch]]:  # pylint: disable=no-self-argument
        if not hasattr(cls, "_name_to_reason"):
            cls._name_to_reason = {name: cls.id_to_reason[i] for name, i in cls.name_to_id.items()}
        return cls._name_to_reason

    # pylint: disable=no-self-argument
    @classproperty
    def db(cls) -> hyperscan.Database:
        if cls._db is None:
            i = 0
            for reason in cls._reasons:
                if not reason.patterns:
                    cls.id_to_reason[i] = reason
                    i += 1
                    continue
                for pattern in reason.patterns:
                    cls.id_to_pattern[i] = pattern
                    cls.id_to_reason[i] = reason
                    i += 1
            cls._reason_to_id = {reason: i for i, reason in cls.id_to_reason.items()}
            cls._name_to_id = {reason.__name__: i for i, reason in cls.id_to_reason.items()}
            if len(cls.reason_to_id) != len(cls.name_to_id):
                raise NotImplementedError()

            ids, expressions = zip(*cls.id_to_pattern.items())
            flags = [cls.id_to_reason[i].pattern_flags for i in ids]
            db = hyperscan.Database()
            db.compile(expressions=expressions, ids=ids, elements=len(expressions), flags=flags)
            cls._db = db
        return cls._db

    @classmethod
    def _match_event_handler(
        cls, id: int, start: int, end: int, flags: int, context: tuple[Job, list[LogFileMatch]]
    ) -> bool | None:
        job, results = context
        reason_cls = cls.id_to_reason[id]
        kwargs = {}
        if issubclass(reason_cls, AppLogFailureReason):
            kwargs["priority"] = id
        results.append(cls.id_to_reason[id](job, start, end, **kwargs))

    @classmethod
    def scan_job_log(cls, job: Job) -> LogFileMatches:
        log_analysis_results = []
        if job.files.steps[-1].application_log is not None:
            cls.db.scan(  # pylint: disable=no-member
                job.files.steps[-1].application_log,
                context=(job, log_analysis_results),
                match_event_handler=cls._match_event_handler,
            )

        problems = []
        log_pos_to_file_open: dict[int, tuple[bool, str]] = {}
        # first_print_per_stream = {}
        # failed_to_open_events = {}
        event_loop_end = sys.maxsize
        application_manager_start = None
        for match in log_analysis_results:
            if isinstance(match, AppLogFailureReason):
                problems.append(match)
            elif isinstance(match, ConnectToFileEvent):
                log_pos_to_file_open[match._start] = (False, match.lfn)
            elif isinstance(match, (EndEventLoopEvent, ApplicationManagerStopEvent)):
                event_loop_end = min(event_loop_end, match._end)
            elif isinstance(match, EventSelectorPrintEvent):
                input_lfns = job.files.steps[-1].prodrun_config["input"]["files"]
                log_pos_to_file_open[match._start] = (False, input_lfns[match.stream_index - 1])
                # if match.stream_index not in first_print_per_stream.values():
                #     first_print_per_stream[match._start] = match.stream_index
            elif isinstance(match, FailedToOpenFileEvent):
                lfn = job.files.xml_catalog.find(f".//File[@ID='{match.guid}']").find(".//logical/lfn").attrib["name"]
                if not lfn.startswith("LFN:"):
                    lfn = f"LFN:{lfn}"
                log_pos_to_file_open[match._start] = (True, lfn)
            elif isinstance(match, ApplicationManagerStartEvent):
                application_manager_start = match._start
            elif isinstance(match, NewEventSelectorPrintEvent):
                log_pos_to_file_open[match._start] = (False, match.lfn)

        if log_pos_to_file_open:
            # Simplify log_pos_to_file_open to only contain the file start positions.
            # For files where we only have end positions, replaced them with a best
            # guess of when that file starts
            to_iter = sorted(log_pos_to_file_open.items())
            (b_idx, (b_end, b_lfn)) = to_iter[0]
            log_pos_to_file_open: dict[int, str] = {application_manager_start + 1 if b_end else b_idx: b_lfn}
            for (a_idx, (a_end, a_lfn)), (b_idx, (b_end, b_lfn)) in pairwise(to_iter):
                if a_end == b_end and a_lfn == b_lfn:
                    continue
                log_pos_to_file_open[a_idx + 1 if b_end else b_idx] = b_lfn

        boundaries = sorted(log_pos_to_file_open)
        result = LogFileMatches(log_pos_to_file_open.values())
        for problem in problems:
            if not boundaries or problem._start < boundaries[0]:
                result.init.append(problem)
            elif problem._start > event_loop_end:
                result.finalize.append(problem)
            else:
                for boundary in boundaries:
                    if problem._start < boundary:
                        break
                    lfn = log_pos_to_file_open[boundary]
                result.by_lfn[lfn].append(problem)

        if log_pos_to_file_open:
            input_lfns = job.files.steps[-1].prodrun_config["input"]["files"]
            last_found_lfn = log_pos_to_file_open[max(log_pos_to_file_open)]
            last_found_lfn_idx = input_lfns.index(last_found_lfn)
            if len(input_lfns) > last_found_lfn_idx + 1:
                next_lfn = input_lfns[last_found_lfn_idx + 1]
                if (
                    job.pre_analysis[last_found_lfn] == "probably_ok"
                    and job.pre_analysis[next_lfn] == "needs_log_analysis"
                ):
                    result.by_lfn[next_lfn] = result.by_lfn.pop(last_found_lfn)
                    result.by_lfn[last_found_lfn] = []

        result.apply_sorting()
        return result

    def __init_subclass__(cls, *, abstract=False) -> None:
        """Register a new failure reason"""
        if cls._db is not None:
            raise TypeError("Cannot register a new reason after the database has been compiled!")

        patterns = None
        if hasattr(cls, "pattern") and hasattr(cls, "patterns"):
            raise TypeError("LogFileMatch subclasses must not have both pattern and patterns!")
        elif hasattr(cls, "pattern"):
            patterns = [cls.pattern]
        elif hasattr(cls, "patterns"):
            patterns = cls.patterns

        if abstract:
            if patterns is not None:
                raise TypeError("LogFileMatch subclasses must not have a pattern or patterns!", cls)
        elif patterns is None:
            raise TypeError("LogFileMatch subclasses must have a pattern or patterns!", cls)
        else:
            cls._reasons.append(cls)
            if not hasattr(cls, "patterns"):
                cls.patterns = patterns

        return super().__init_subclass__()


class AppLogFailureReason(LogFileMatch, abstract=True):
    """Describes a part of the application log which indicates a problem."""

    def __init__(self, job: Job, start: int, end: int, *, priority: int) -> None:
        super().__init__(job, start, end)
        self.priority = priority

    def __len__(self) -> int:
        return self._end - self._start

    def __lt__(self, other):
        return (self.priority, -len(self)) < (other.priority, -len(other))


class RecoverableFailure(AppLogFailureReason, abstract=True): ...


class RecoverableFailureWithManualIntervention(RecoverableFailure, abstract=True): ...


class KnownUnrecoverableFailure(AppLogFailureReason, abstract=True):
    issue_url: str


class WorkerNodeIssue(RecoverableFailure, abstract=True): ...


class EventTimeout(AppLogFailureReason, abstract=True): ...


class UnknownFailure(AppLogFailureReason, abstract=True): ...


class Event(LogFileMatch, abstract=True):
    """Describes a part of the application log which indicates an event like connecting to a file."""


class ConnectToFileEvent(Event):
    pattern = rb"(?:RootIOAlg|Stream:EventSelector).+DATAFILE='(LFN:[^']+)'"
    RE_PATTERN = re.compile(pattern)

    @property
    def lfn(self) -> str:
        return self.RE_PATTERN.search(self.match_string.encode()).group(1).decode()


class DisconnectFromFileEvent(Event):
    pattern = rb"Disconnect from dataset.+\n"


class EndEventLoopEvent(Event):
    pattern = rb"No more events in event selection"


class EventSelectorPrintEvent(Event):
    pattern = rb"Record number within stream (\d+):"
    RE_PATTERN = re.compile(pattern)

    @property
    def stream_index(self) -> int:
        return int(self.RE_PATTERN.search(self.match_string.encode()).group(1))


class NewEventSelectorPrintEvent(Event):
    pattern = rb"Reading Event record (\d+) within ([^\s]+?),?\s"
    RE_PATTERN = re.compile(pattern)

    @property
    def stream_index(self) -> int:
        return int(self.RE_PATTERN.search(self.match_string.encode()).group(1))

    @property
    def lfn(self) -> str:
        return self.RE_PATTERN.search(self.match_string.encode()).group(2).decode()


class FailedToOpenFileEvent(Event):
    pattern = rb"Failed to open dsn:([^ ]+) "
    RE_PATTERN = re.compile(pattern)

    @property
    def guid(self) -> str:
        return self.RE_PATTERN.search(self.match_string.encode()).group(1).decode()


class ApplicationManagerStartEvent(Event):
    pattern = rb"Application Manager Started successfully"


class ApplicationManagerStopEvent(Event):
    pattern = rb"Application Manager Stopped successfully"



###############################################################################
# (c) Copyright 2025 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import json
from datetime import datetime, timezone
from urllib.parse import quote_plus

from collections import defaultdict

from sqlalchemy import create_engine, Column, Integer, String, BigInteger, Boolean, text, delete, Engine, select, func
from sqlalchemy.orm import declarative_base
from sqlalchemy.sql.expression import and_, false, true

from DIRAC.ConfigurationSystem.Client.Utilities import getDBParameters
from DIRAC.Core.Base.DIRACDB import DIRACDB
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise

# Define the SQLAlchemy model
Base = declarative_base()

LOGICAL_SE_NAME = "FakeSE"


class StorageSummary(Base):
    __tablename__ = "dump_storage_summary"

    SEName = Column(String(255), primary_key=True)
    SESize = Column(BigInteger)
    SEFiles = Column(BigInteger)
    SESite = Column(String(255))
    SEType = Column(String(255))
    IsDisk = Column(Boolean)


# Some folders are "duplicated", with case sensitivity issue
# PHYSICS / physics
# ALICE / alice
# +----------------------------------------+---+
# | Name                                   | t |
# +----------------------------------------+---+
# | /lhcb/data/2008/RAW/LHCb/PHYSICS/21188 | 2 |
# | /lhcb/data/2008/RAW/LHCb/PHYSICS/21192 | 2 |
# | /lhcb/data/2008/RAW/RICH2/ALICE/21933  | 2 |
# | /lhcb/data/2008/RAW/RICH2/ALICE/21936  | 2 |
# | /lhcb/data/2008/RAW/RICH2/ALICE/21937  | 2 |
# | /lhcb/data/2008/RAW/RICH2/ALICE/21938  | 2 |
# +----------------------------------------+---+
# that's why we need the specific mysql collate


class DirectoryMetadata(Base):
    __tablename__ = "dump_directory_metadata"

    Name = Column(String(255), primary_key=True)
    production = Column(Integer)
    filetype = Column(String(255))
    online_stream = Column(String(255))
    maybe_eventtype = Column(Integer)
    ConfigName = Column(String(255))
    ConfigVersion = Column(String(255))
    Description = Column(String(255))
    ProcPath = Column(String(255))
    EventTypeID = Column(Integer)

    __table_args__ = {"mysql_engine": "InnoDB", "mysql_charset": "utf8mb4", "mysql_collate": "utf8mb4_bin"}


class UserDirectoryMetadata(Base):
    """Table to store information about the users space consumption"""

    __tablename__ = "dump_user_directory_metadata"

    User = Column(String(255), primary_key=True)
    SEName = Column(String(255), primary_key=True)
    SESize = Column(BigInteger)
    SEFiles = Column(BigInteger)
    UserStatus = Column(String(255))


class StorageUsageDumpDB(DIRACDB):

    def __init__(self, *, url=None, parentLogger=None):
        self.fullname = self.__class__.__name__
        super().__init__(parentLogger=parentLogger)
        if url is None:
            param = returnValueOrRaise(getDBParameters("DataManagement/StorageUsageDB"))
            url = f"mysql://{param['User']}:{quote_plus(param['Password'])}@{param['Host']}:{param['Port']}/{param['DBName']}"

        self.engine = create_engine(
            url,
            echo=False,
            pool_recycle=30 * 60,
        )
        Base.metadata.create_all(self.engine)

    def directory_metadata_to_sql(self, df):
        with self.engine.connect() as conn:
            conn.execute(delete(DirectoryMetadata))

            df.to_sql(
                DirectoryMetadata.__tablename__,
                conn,
                index=False,
                method="multi",
                chunksize=10000,
                if_exists="append",
            )
            new_comment = json.dumps({"LastUpdate": datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")})
            conn.execute(text(f"ALTER TABLE {DirectoryMetadata.__tablename__} COMMENT = '{new_comment}';"))

    def _get_last_update(self, table_name: str) -> datetime:
        with self.engine.connect() as conn:
            comment = conn.execute(text(f"SHOW TABLE STATUS WHERE name = '{table_name}';")).one()[-1]
            return datetime.strptime(json.loads(comment)["LastUpdate"], "%Y-%m-%d %H:%M:%S")

    def storage_summary_to_sql(self, df):
        with self.engine.connect() as conn:
            conn.execute(delete(StorageSummary))
            df.to_sql(
                StorageSummary.__tablename__,
                conn,
                index=False,
                method="multi",
                chunksize=10000,
                if_exists="append",
            )
            new_comment = json.dumps({"LastUpdate": datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")})
            conn.execute(text(f"ALTER TABLE {StorageSummary.__tablename__} COMMENT = '{new_comment}';"))

    def get_site_usage(self, site_name: str, *, tape: bool = True, disk: bool = True) -> tuple[int, datetime]:
        """
        Retrieves the total storage usage and the last update time for a given site.

        Parameters:
        site_name (str): The name of the site for which to retrieve the storage usage.
        tape (bool): Flag indicating whether to include tape storage in the calculation. Default is True.
        disk (bool): Flag indicating whether to include disk storage in the calculation. Default is True.

        Returns:
        tuple[int, datetime]: A tuple containing the total storage size in bytes and the last update time.
        """
        with self.engine.connect() as conn:
            whereConditions = [StorageSummary.SESite == site_name]
            if tape and not disk:
                whereConditions.append(StorageSummary.IsDisk == false())
            elif disk and not tape:
                whereConditions.append(StorageSummary.IsDisk == true())
            query = select(func.sum(StorageSummary.SESize)).where(and_(*whereConditions))
            size_scal = conn.execute(query).scalar()
            if size_scal is None:
                raise ValueError(f"No such site {site_name}")
            size = int(size_scal)
            last_update = self._get_last_update(StorageSummary.__tablename__)
            # Execute the query
            return (size, last_update)

    def get_se_usage(self, se_name: str) -> tuple[int, datetime]:
        """
        Retrieves the total storage usage and the last update time for a given site.

        Parameters:
        site_name (str): The name of the se for which to retrieve the storage usage.

        Returns:
        tuple[int, datetime]: A tuple containing the total storage size in bytes and the last update time.
        """
        with self.engine.connect() as conn:
            query = select(StorageSummary.SESize).where(StorageSummary.SEName == se_name)
            size_scal = conn.execute(query).scalar()
            if size_scal is None:
                raise ValueError(f"No such se {se_name}")
            size = int(size_scal)
            last_update = self._get_last_update(StorageSummary.__tablename__)
            return (size, last_update)

    def user_storage_to_sql(self, df):
        with self.engine.connect() as conn:
            conn.execute(delete(UserDirectoryMetadata))

            df.to_sql(
                UserDirectoryMetadata.__tablename__,
                conn,
                index=False,
                method="multi",
                chunksize=1000,
                if_exists="append",
            )
            new_comment = json.dumps({"LastUpdate": datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S")})
            conn.execute(text(f"ALTER TABLE {UserDirectoryMetadata.__tablename__} COMMENT = '{new_comment}';"))

    def get_user_storage_summary(self) -> tuple[dict[str, int], datetime]:
        """Get the total physical size used by each user"""
        with self.engine.connect() as conn:
            query = (
                select(UserDirectoryMetadata.User, func.sum(UserDirectoryMetadata.SESize))
                .where(UserDirectoryMetadata.SEName != LOGICAL_SE_NAME)
                .group_by(UserDirectoryMetadata.User)
            )

            user_summary = {user: int(size) for user, size in conn.execute(query).all()}

            last_update = self._get_last_update(UserDirectoryMetadata.__tablename__)
            return (user_summary, last_update)

    def get_user_logical_summary(self, username: str = "") -> tuple[dict[str, dict[str, int]], datetime]:
        """Get the total physical size used by each user"""
        with self.engine.connect() as conn:
            query = select(
                UserDirectoryMetadata.User, UserDirectoryMetadata.SESize, UserDirectoryMetadata.SEFiles
            ).where(UserDirectoryMetadata.SEName == LOGICAL_SE_NAME)

            if username:
                query = query.where(UserDirectoryMetadata.User == username)

            user_summary = {}

            for username, size, files in conn.execute(query).all():
                user_summary[username] = {"Size": size, "Files": files}

            last_update = self._get_last_update(UserDirectoryMetadata.__tablename__)
            return (user_summary, last_update)

    def get_user_storage_summary_per_se(
        self, username: str = ""
    ) -> tuple[dict[str, dict[str, dict[str, int]]], datetime]:
        """Get the physical size per user per se."""
        with self.engine.connect() as conn:
            query = select(
                UserDirectoryMetadata.User,
                UserDirectoryMetadata.SEName,
                UserDirectoryMetadata.SESize,
                UserDirectoryMetadata.SEFiles,
            ).where(UserDirectoryMetadata.SEName != LOGICAL_SE_NAME)

            if username:
                query = query.where(UserDirectoryMetadata.User == username)

            user_summary_per_se = defaultdict(dict)

            for username, se_name, size, files in conn.execute(query).all():
                user_summary_per_se[username][se_name] = {"Size": size, "Files": files}

            last_update = self._get_last_update(UserDirectoryMetadata.__tablename__)

            return (user_summary_per_se, last_update)



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

__all__ = [
    "read_zip",
    "file_exists",
]

import asyncio
import io
from zipfile import ZipFile

import pyxrootd.client

from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise


async def _get_xrd_file(log_lfn: str) -> pyxrootd.client.File | None:
    """Get a pyxrootd file object for the given LFN.

    This function will first try to open the file directly from CERN, and if that fails
    it will try to get the PFN from DIRAC and open that. This might happen if the file
    is not yet available on CERN's EOS and is still in FAILOVER.
    """
    loop = asyncio.get_running_loop()

    f = pyxrootd.client.File()
    status, _ = await loop.run_in_executor(
        None, f.open, f"root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/logSE/{log_lfn}"
    )
    if status["ok"]:
        return f

    from DIRAC.DataManagementSystem.Client.DataManager import DataManager

    result = await loop.run_in_executor(None, DataManager().getReplicas, log_lfn)
    if result["OK"] and (log_pfns := result["Value"]["Successful"].get(log_lfn)):
        for log_pfn in log_pfns.values():
            print("Did not find log file at CERN, trying FAILOVER:", log_pfn)
            f = pyxrootd.client.File()
            status, _ = await loop.run_in_executor(None, f.open, log_pfn)
            if status["ok"]:
                return f


async def read_zip(job_id: int, log_lfn: str) -> dict[str, bytes]:
    loop = asyncio.get_running_loop()
    cancelled = False

    def handle_read(status, data: bytes, host_list):
        if not cancelled:
            loop.call_soon_threadsafe(queue.put_nowait, (status, data))

    queue = asyncio.Queue()
    f = await _get_xrd_file(log_lfn)
    if f is None:
        print("Failed to open log file:", log_lfn)
        return job_id, {}

    status = f.read(callback=handle_read)
    if not status["ok"]:
        raise NotImplementedError(log_lfn, status)

    status, data = await asyncio.wait_for(queue.get(), 30)
    if not status["ok"]:
        raise NotImplementedError(log_lfn, status)

    status, _ = await loop.run_in_executor(None, f.close)
    if not status["ok"]:
        raise NotImplementedError(log_lfn, status)

    zf = ZipFile(io.BytesIO(data))
    files = {}
    for zi in zf.infolist():
        with zf.open(zi) as f:
            # Only read the first 100 MiB to avoid large files taking too long
            if zi.file_size > 100 * 1024**2:
                print(
                    f"Warning file for {job_id} is {zi.file_size / 1024**2:.1f} MiB, "
                    f"only reading first and last 10 MiB from {log_lfn}:{zi.filename}"
                )
                files[zi.filename.split("/", 1)[1]] = f.read(10 * 1024**2)
                files[zi.filename.split("/", 1)[1]] += b"\nTRUNCATED\n"
                f.seek(zi.file_size - 10 * 1024**2)
                files[zi.filename.split("/", 1)[1]] += f.read()
            else:
                files[zi.filename.split("/", 1)[1]] = f.read()

    return job_id, files


async def file_exists(log_lfn: str) -> bool:
    """Check if the file exists on the XRootD server."""
    f = await _get_xrd_file(log_lfn)
    if f is not None:
        await asyncio.get_running_loop().run_in_executor(None, f.close)
        return True
    return False



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import asyncio
import base64
import os
import ssl
import tempfile
from collections import defaultdict
from contextlib import asynccontextmanager
from itertools import chain
from typing import Annotated, Union, Literal

import httpx
import zstandard
from pydantic import RootModel, BaseModel, Field, ValidationError, ConfigDict
from rich.prompt import Prompt, Confirm
from rich import print

from DIRAC import gConfig
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks
from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import removeReplicas
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from LHCbDIRAC.TransformationSystem.Utilities.PluginUtilities import addFilesToTransformation

from .gitlab import APRepoIssue
from ..LogAnalysis.prod_analyzer import GroupedProblems, AggregatedLFNDebugInfo
from ..LogAnalysis.utils import call_dirac
from ..LogAnalysis.output import make_problems_table


def clean_lfns(lfns: list[str]) -> list[str]:
    return [lfn[4:] if lfn.startswith("LFN:") else lfn for lfn in lfns]


def get_all_subclasses(cls):
    subclasses = set()
    work = [cls]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.add(child)
                work.append(child)
    return tuple(subclasses)


def build_action_union():
    subclasses = get_all_subclasses(BaseAction)
    return Union[tuple(subclasses)]


class BaseAction(BaseModel):
    model_config = ConfigDict(extra="forbid")

    action_name: str
    _log_lines: list[str] = []

    def __init_subclass__(cls):
        super().__init_subclass__()
        cls.__annotations__["action_name"] = f"{Literal.__name__}[{cls.__name__!r}]"
        cls.action_name = cls.__name__
        cls.model_rebuild()

    async def run(self, overview):
        raise NotImplementedError()

    def log(self, message):
        self._log_lines.append(message)
        print(f"{self.__class__.__name__}:{message}")


class BaseTransformAction(BaseAction):
    transform_id: int

    def log(self, message):
        self._log_lines.append(message)
        print(f"{self.transform_id}:{self.__class__.__name__}:{message}")


class ReduceGroupSizeAction(BaseTransformAction):

    async def run(self, overview):
        self.log(f"Reducing group size for {self.transform_id}")
        await call_dirac(TransformationClient().setTransformationParameter, self.transform_id, "GroupSize", 1)


class SetDownloadInputDataAction(BaseTransformAction):

    async def run(self, overview):
        by_protocol_plugin_name = "DIRAC.WorkloadManagementSystem.Client.InputDataByProtocol"
        download_plugin_name = "DIRAC.WorkloadManagementSystem.Client.DownloadInputData"

        self.log(f"Setting download input data for {self.transform_id}")
        body = await call_dirac(TransformationClient().getTransformationParameters, self.transform_id, "Body")
        if download_plugin_name not in body:
            if body.count(by_protocol_plugin_name) != 1:
                raise NotImplementedError(self)
            body = body.replace(by_protocol_plugin_name, download_plugin_name)
            await call_dirac(TransformationClient().setTransformationParameter, self.transform_id, "Body", body)


class ResetFileStatusAction(BaseTransformAction):
    transform_id: int
    lfns_to_reset: Annotated[set[str], Field(min_length=1)]
    new_status: str = "Unused"

    async def run(self, overview):
        lfns = sorted(set(clean_lfns(self.lfns_to_reset)))
        self.log(f"Resetting {len(lfns)} in {self.transform_id}")
        query = {
            "TransformationID": self.transform_id,
            "LFN": lfns,
        }
        result = await call_dirac(
            TransformationClient().getTransformationFiles, condDict=query, columns=["TransformationID", "LFN", "Status"]
        )
        if {x["LFN"] for x in result} != set(lfns):
            raise NotImplementedError(self, result)
        if not all(x["Status"] in {"MaxReset", "Problematic"} for x in result):
            raise NotImplementedError(self, result)
        result = await call_dirac(
            TransformationClient().setFileStatusForTransformation, self.transform_id, self.new_status, lfns, force=True
        )
        for lfn, lfn_new_status in result.items():
            self.log(f"Set status for {lfn} to {lfn_new_status}")
        if not_updated := set(lfns) - set(result):
            raise NotImplementedError(self, not_updated)


class PostAPIssueDiscussion(BaseAction):
    ap_key: tuple[str, str, str]
    body: Annotated[str, Field(max_length=900_000)]
    attachments: dict[str, str] = {}

    async def run(self, overview):
        ap = overview.key_to_ap[self.ap_key]
        issue = APRepoIssue(overview.repo.get_issue(ap.issue_url))
        format_kwargs = {}
        for name, content in self.attachments.items():
            uploaded_file = overview.repo.project.upload(name, filedata=content)
            format_kwargs[name.replace(".", "_")] = uploaded_file["url"]
        if format_kwargs:
            self.body = self.body.format(**format_kwargs)
        discussion = issue.discussions.create({"body": self.body})
        self.log(f"Posted discussion to {ap.issue_url} as {discussion.get_id()}")


class RequestFeedbackAction(BaseTransformAction):
    description: str
    instances: list[AggregatedLFNDebugInfo]
    reset_action: AnyAction
    abort_action: AnyAction

    async def run(self, overview):
        request_id = overview.transform_id_to_request_id[self.transform_id]
        aps = overview.request_id_to_aps[request_id]
        ap = sorted(aps, key=lambda x: x.created)[0]
        issue = APRepoIssue(overview.repo.get_issue(ap.issue_url))
        lfns = sorted([x.lfn for x in self.instances])

        for discussion_id, compressed_info in issue.metadata.open_discussions.items():
            info = RequestFeedbackAction.model_validate_json(zstandard.decompress(base64.b64decode(compressed_info)))
            if info.transform_id != self.transform_id:
                continue
            info_lfns = {x.lfn for x in info.instances}
            if not info_lfns.issubset(lfns):
                continue
            if info_lfns != set(lfns):
                raise NotImplementedError(self, issue, info_lfns, lfns)
            await self._prompt_for_action(overview, issue, discussion_id)
            break
        else:
            self._post_discussion(ap, issue, lfns)

    def _post_discussion(self, ap, issue, lfns):
        body = []
        body.append("Please investigate the following problems and decide if you")
        body.append("would like to skip processing these files.")
        body.append("")
        body.append(self.description)
        body.append("")
        body.append("<details><summary>Click to show details</summary>")
        body.append("")
        body.extend(make_problems_table(self.instances, is_summary=True))
        body.append("")
        body.append("</details>")
        body = "\n".join(body)
        discussion = issue.discussions.create({"body": body})
        self.log(f"Posted discussion to {ap.issue_url} as {discussion.get_id()}")

        issue.labels.append("Needs investigation")
        issue.metadata.open_discussions[discussion.get_id()] = base64.b64encode(
            zstandard.compress(self.model_dump_json().encode())
        ).decode()
        issue.update_metadata()

    async def _prompt_for_action(self, overview, issue, discussion_id):
        note_id = issue.discussions.get(discussion_id).attributes["notes"][0]["id"]
        print(f"Discussion at {issue.url}#note_{note_id}")
        match choice := Prompt.ask(
            "What would you like to do?", choices=["Reset", "Abort", "Nothing"], default="Nothing"
        ):
            case "Reset":
                print("Running reset action")
                await self.reset_action.run(overview)
            case "Abort":
                print("Running abort action")
                await self.abort_action.run(overview)
            case "Nothing":
                return
            case _:
                raise NotImplementedError(choice)

        issue.metadata.open_discussions.pop(discussion_id)
        if not issue.metadata.open_discussions:
            issue.labels.remove("Needs investigation")
        issue.update_metadata()


class SetTransformStateAction(BaseTransformAction):
    state: Literal["Stopped"] | Literal["Completed"] | Literal["Flush"]

    async def run(self, overview):
        self.log(f"Setting state to {self.state}")
        await call_dirac(
            TransformationClient().setTransformationParameter, self.transform_id, "Status", self.state, force=True
        )


class AddLFNsToReplicationTransform(BaseAction):
    lfns: list[str]
    se_name: str

    async def run(self, overview):
        self.log(f"Adding {len(self.lfns)} LFNs to {self.se_name}")
        transform_name = f"Replicate-to-{self.se_name}"
        replication_transform = returnValueOrRaise(TransformationClient().getTransformation(transform_name))
        replication_transform_id = replication_transform["TransformationID"]
        result = returnValueOrRaise(
            addFilesToTransformation(replication_transform_id, self.lfns, addRunInfo=True, resetPresentFiles=True)
        )
        if result["Failed"]:
            raise NotImplementedError(self, result["Failed"])
        self.log(f"Added {len(self.lfns)} LFNs to {transform_name} ({replication_transform_id}): {self.lfns}")


class HandleInputDataResolutionFailureAction(BaseAction):
    transform_id_to_lfns: dict[int, list[str]]

    async def run(self, overview):
        # Prompt for if this should be ran
        if not Confirm.ask(f"Do input data resolution checks for {len(self.transform_id_to_lfns)} transforms?"):
            self.log("Skipping handling input data resolution failure")
            return

        lfns = clean_lfns(sorted(set(chain(*self.transform_id_to_lfns.values()))))
        self.log(f"Handling input data resolution failure for {len(lfns)} LFNs in {sorted(self.transform_id_to_lfns)}")
        cc = ConsistencyChecks()
        cc.lfns = lfns
        cc.checkFC2SE(bkCheck=True)

        lost_lfns = set()
        post_logbook = False
        okay_lfns = set(lfns)

        if cc.notRegisteredAtSE:
            okay_lfns -= set(cc.notRegisteredAtSE)
            raise NotImplementedError(self)

        if cc.existLFNsBKRepNo:
            okay_lfns -= set(cc.existLFNsBKRepNo)
            raise NotImplementedError(self)

        if cc.existLFNsNotInBK:
            okay_lfns -= set(cc.existLFNsNotInBK)
            raise NotImplementedError(self)

        if cc.existLFNsBadFiles or cc.existLFNsNotExisting:
            okay_lfns -= set(cc.existLFNsBadFiles) | set(cc.existLFNsNotExisting)
            raise NotImplementedError(self)

        if cc.existLFNsBadReplicas:
            okay_lfns -= set(cc.existLFNsBadReplicas)
            raise NotImplementedError(self)

        if cc.existLFNsNoSE:
            okay_lfns -= set(cc.existLFNsNoSE)
            by_se = defaultdict(set)
            # Group them by SE
            for lfn, ses in cc.existLFNsNoSE.items():
                for se in ses:
                    if se == "All":
                        raise NotImplementedError(self)
                    by_se[se].add(lfn)
            # Remove the replicas
            for se, se_lfns in by_se.items():
                se_lfns = sorted(se_lfns)
                self.log(f"Removing {len(se_lfns)} LFNs from {se}: {se_lfns}")
                return_code, error_reasons = removeReplicas(se_lfns, [se])
                if return_code != 0:
                    raise NotImplementedError(self, return_code, error_reasons)
                if error_reasons:
                    raise NotImplementedError(self, return_code, error_reasons)
                await AddLFNsToReplicationTransform(se_name=se, lfns=se_lfns).run(overview)
            post_logbook = True

        # If we have some LFNs that seem to be okay, double-check by downloading
        # a small part of the file to make sure the server containing the file
        # is actually reachable and the file is readable.
        if okay_lfns:
            from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
            from DIRAC.Core.Security.Locations import getCAsLocation, getProxyLocation
            from DIRAC.DataManagementSystem.Client.DataManager import DataManager

            replicas = returnValueOrRaise(DataManager().getReplicas(lfns, diskOnly=True, protocol="https"))
            if replicas["Failed"]:
                raise NotImplementedError(f"Failed to get some replicas {replicas['Failed']}")

            ctx = ssl.create_default_context(purpose=ssl.Purpose.SERVER_AUTH, capath=getCAsLocation())
            ctx.load_cert_chain(getProxyLocation())

            errored_lfns = dict()
            tasks = defaultdict(list)
            # When storage is misbehaving, checking replicas can be extremely
            # slow due to timeouts so do the checks in parallel.
            async with httpx.AsyncClient(timeout=90, verify=ctx, follow_redirects=True) as client:
                async with asyncio.TaskGroup() as tg:
                    for lfn, lfn_replicas in replicas["Successful"].items():
                        assert len(lfn_replicas) > 0
                        any_replica_okay = False
                        for se, pfn in lfn_replicas.items():
                            coro = check_replica_worker(self.log, client, lfn, pfn, se, errored_lfns)
                            tasks[lfn].append(tg.create_task(coro))

            # Gather the results and see if any replica was okay for each LFN
            for lfn, lfn_replicas in replicas["Successful"].items():
                any_replica_okay = any(t.result() for t in tasks[lfn])
                if not any_replica_okay:
                    okay_lfns.remove(lfn)
            if errored_lfns and Confirm.ask(
                f"Some LFNs had all replicas unreachable ({len(errored_lfns)} SEs). Make logbook entry?"
            ):
                subject = "Consistency checks okay but replicas unreachable"
                body = "The following LFNs had all replicas unreachable:\n\n"
                for se, se_lfns in errored_lfns.items():
                    body += f"### {se}\n\n"
                    for lfn, (code, reason) in se_lfns.items():
                        body += f"- {lfn}: {code} {reason[:200]!r}\n"
                    body += "\n"
                body += f"\n\n{self.model_dump_json(indent=2)}"
                overview.logbook.create_post(
                    "", list(self.transform_id_to_lfns), subject, body, {}, ["Data Management", "WGP"]
                )

        lfns_to_skip = {f"LFN:{lfn}" for lfn in sorted(chain(*errored_lfns.values()))}
        for transform_id, transform_lfns in self.transform_id_to_lfns.items():
            transform_lfns = sorted(set(transform_lfns) - lost_lfns - lfns_to_skip)
            if transform_lfns:
                self.log(f"Resetting {len(transform_lfns)} LFNs Unused in {transform_id}: {transform_lfns}")
                await ResetFileStatusAction(transform_id=transform_id, lfns_to_reset=transform_lfns).run(overview)

        if post_logbook:
            subject = "Handling InputDataResolutionFailure for AP"
            body = "\n".join(self._log_lines)
            body += f"\n\n{self.model_dump_json(indent=2)}"
            overview.logbook.create_post(
                "", list(self.transform_id_to_lfns), subject, body, {}, ["Data Management", "WGP"]
            )


async def check_replica_worker(log, client, lfn, pfn, se, errored_lfns):
    try:
        async with client.stream("GET", pfn, headers={"Range": "bytes=0-32"}) as response:
            if response.status_code >= 400:
                log(f"Error {response.status_code} from {pfn}: {await response.aread()}")
                errored_lfns.setdefault(se, {})[lfn] = (response.status_code, await response.aread())
                return False
            total = int(response.headers.get("Content-Length", 0))
            log(f"Got {total} bytes from {pfn}")
            data = await response.aread()
            assert data
            return True
    except Exception as e:
        log(f"Exception from {pfn}: {e!r}")
        errored_lfns.setdefault(se, {})[lfn] = (type(e).__name__, str(e))
        return False


@asynccontextmanager
async def get_admin_proxy_path():
    with tempfile.NamedTemporaryFile("w") as f:
        while True:
            if hasattr(get_admin_proxy_path, "__password"):
                password = get_admin_proxy_path.__password
            else:
                password = Prompt.ask("Enter certificate password", password=True)
            cmd = ["dirac-proxy-init", "-g", "lhcb_admin", "--pwstdin", "--out", f.name]
            print(f"Creating admin proxy using {cmd}")
            proc = await asyncio.create_subprocess_exec(*cmd, stdin=asyncio.subprocess.PIPE)
            await proc.communicate(password.encode())
            if proc.returncode == 0:
                break
            print("Failed to create proxy, try again")
            delattr(get_admin_proxy_path, "__password")
        yield f.name
    get_admin_proxy_path.__password = password


class SendToHospital(BaseAction):
    transform_ids: list[int]

    async def run(self, overview):
        await call_dirac(gConfig.forceRefresh)

        cs_section_name = "/Operations/Defaults/Hospital/Clinics/AP/Transformations"
        original_tids = set(map(int, returnValueOrRaise(gConfig.getOption(cs_section_name, []))))
        new_tids = sorted(original_tids | set(self.transform_ids))
        if set(new_tids) == original_tids:
            return
        async with get_admin_proxy_path() as proxy_path:
            proc = await asyncio.create_subprocess_exec(
                "lb-dirac",
                "dirac-configuration-cli",
                stdin=asyncio.subprocess.PIPE,
                env=os.environ | {"X509_USER_PROXY": proxy_path},
            )
            commands = [
                f"set {cs_section_name} {','.join(map(str, new_tids))}",
                "showDiffWithServer",
                "writeToServer",
                "yes",
            ]
            await proc.communicate("\n".join(commands).encode())
            if proc.returncode != 0:
                raise NotImplementedError(self, proc.returncode)


AnyAction = Annotated[build_action_union(), Field(discriminator="action_name")]


class ActionsList(RootModel):
    root: dict[str, list[AnyAction]]

    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    def __len__(self):
        return len(self.root)

    def items(self):
        return self.root.items()



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import pytest

from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.classes import (
    JobArtifacts,
    JobArtifactConsistencyError,
)

from . import load_example


@pytest.mark.parametrize(
    "in_files, out_keys, out_steps_keys",
    [
        ({}, {}, []),
        (
            {
                "DaVinci_00244431_00000106_1.log": b"DAVINCI_00244431_00000106_1.LOG",
                "jobDescription.xml": b"JOBDESCRIPTION.XML",
                "job.info": b"JOB.INFO",
                "pool_xml_catalog.xml": b"POOL_XML_CATALOG.XML",
                "prmon_00244431_00000106_1.txt": b"PRMON_00244431_00000106_1.TXT",
                "prmon.log": b"PRMON.LOG",
                "prodConf_00244431_00000106_1.py": b"PRODCONF_00244431_00000106_1.PY",
                "prodConf_DaVinci_00244431_00000106_1.json": b"PRODCONF_DAVINCI_00244431_00000106_1.JSON",
                "std.out": b"STD.OUT",
                "summaryDaVinci_00244431_00000106_1.xml": b"SUMMARYDAVINCI_00244431_00000106_1.XML",
            },
            {
                "jobDescription.xml",
                "job.info",
                "pool_xml_catalog.xml",
                "prmon.log",
                "std.out",
            },
            [
                {
                    "DaVinci_00244431_00000106_1.log",
                    "prmon_00244431_00000106_1.txt",
                    "prodConf_00244431_00000106_1.py",
                    "prodConf_DaVinci_00244431_00000106_1.json",
                    "summaryDaVinci_00244431_00000106_1.xml",
                }
            ],
        ),
        (
            {
                "Moore_00244431_00000106_2.log": b"MOORE_00244431_00000106_2.LOG",
                "Boole_00244431_00000106_1.log": b"BOOLE_00244431_00000106_1.LOG",
                "bookkeeping_00244431_00000106_1.xml": b"BOOKKEEPING_00244431_00000106_1.XML",
                "bookkeeping_00244431_00000106_2.xml": b"BOOKKEEPING_00244431_00000106_2.XML",
                "job.info": b"JOB.INFO",
                "jobDescription.xml": b"JOBDESCRIPTION.XML",
                "pool_xml_catalog.xml": b"POOL_XML_CATALOG.XML",
                "prmon.log": b"PRMON.LOG",
                "prmon_00244431_00000106_1.txt": b"PRMON_00244431_00000106_1.TXT",
                "prmon_00244431_00000106_2.txt": b"PRMON_00244431_00000106_2.TXT",
                "prodConf_00244431_00000106_1.py": b"PRODCONF_00244431_00000106_1.PY",
                "prodConf_Boole_00244431_00000106_1.json": b"PRODCONF_BOOLE_00244431_00000106_1.JSON",
                "prodConf_Moore_00244431_00000106_2.json": b"PRODCONF_MOORE_00244431_00000106_2.JSON",
                "std.out": b"STD.OUT",
                "summaryBoole_00244431_00000106_1.xml": b"SUMMARYBOOLE_00244431_00000106_1.XML",
                "summaryMoore_00244431_00000106_2.xml": b"SUMMARYMOORE_00244431_00000106_2.XML",
            },
            {
                "std.out",
                "job.info",
                "jobDescription.xml",
                "pool_xml_catalog.xml",
                "prmon.log",
            },
            [
                {
                    "Boole_00244431_00000106_1.log",
                    "bookkeeping_00244431_00000106_1.xml",
                    "prmon_00244431_00000106_1.txt",
                    "prodConf_00244431_00000106_1.py",
                    "prodConf_Boole_00244431_00000106_1.json",
                    "summaryBoole_00244431_00000106_1.xml",
                },
                {
                    "Moore_00244431_00000106_2.log",
                    "bookkeeping_00244431_00000106_2.xml",
                    "prmon_00244431_00000106_2.txt",
                    "prodConf_Moore_00244431_00000106_2.json",
                    "summaryMoore_00244431_00000106_2.xml",
                },
            ],
        ),
    ],
)
def test_artifact_loading(in_files, out_keys, out_steps_keys):
    # Add the expected output keys
    out_job_files = {k: in_files[k] for k in out_keys}
    out_steps_files = [{k: in_files[k] for k in v} for v in out_steps_keys]
    # Parse the file list
    job_artifacts = JobArtifacts.from_files(in_files, transform_id=244431, task_id=106)
    # Check the output is as expected
    assert out_job_files == job_artifacts._files
    assert out_steps_files == [step._files for step in job_artifacts.steps]


def test_inconsistent_artifact_loading():
    files = {
        "Moore_00241881_00001234_2.log": b"MOORE_00241881_00001234_2.LOG",
        "Boole_00241881_00001234_1.log": b"BOOLE_00241881_00001234_1.LOG",
        "bookkeeping_00241881_00001234_1.xml": b"BOOKKEEPING_00241881_00001234_1.XML",
        "bookkeeping_00241881_00001234_2.xml": b"BOOKKEEPING_00241881_00001234_2.XML",
        "job.info": b"JOB.INFO",
        "jobDescription.xml": b"JOBDESCRIPTION.XML",
        "pool_xml_catalog.xml": b"POOL_XML_CATALOG.XML",
        "prmon.log": b"PRMON.LOG",
        "prmon_00241881_00001234_1.txt": b"PRMON_00241881_00001234_1.TXT",
        "prmon_00241881_00001234_2.txt": b"PRMON_00241881_00001234_2.TXT",
        "prodConf_00241881_00001234_1.py": b"PRODCONF_00241881_00001234_1.PY",
        "prodConf_Boole_00241881_00001234_1.json": b"PRODCONF_BOOLE_00241881_00001234_1.JSON",
        "prodConf_Moore_00241881_00001234_2.json": b"PRODCONF_MOORE_00241881_00001234_2.JSON",
        "std.out": b"STD.OUT",
        "summaryBoole_00241881_00001234_1.xml": b"SUMMARYBOOLE_00241881_00001234_1.XML",
        "summaryMoore_00241881_00001234_2.xml": b"SUMMARYMOORE_00241881_00001234_2.XML",
    }
    # Make sure this works as expected
    JobArtifacts.from_files(files, transform_id=241881, task_id=1234)
    # Check the various error cases
    with pytest.raises(JobArtifactConsistencyError, match="Transformation ID mismatch"):
        JobArtifacts.from_files(files, transform_id=241882, task_id=1234)
    with pytest.raises(JobArtifactConsistencyError, match="Task ID mismatch"):
        JobArtifacts.from_files(files, transform_id=241881, task_id=106)
    bad_files = files | {"prodConf_Boole_12345678_00001234_1.json": b"AAA"}
    with pytest.raises(JobArtifactConsistencyError, match="Transformation ID mismatch"):
        JobArtifacts.from_files(bad_files, transform_id=241881, task_id=1234)
    bad_files = files | {"prodConf_Boole_00241881_00000001_1.json": b"AAA"}
    with pytest.raises(JobArtifactConsistencyError, match="Task ID mismatch"):
        JobArtifacts.from_files(bad_files, transform_id=241881, task_id=1234)


def test_parsing():
    transform_id, task_id, files = load_example(957694667)
    job_artifacts = JobArtifacts.from_files(files, transform_id=transform_id, task_id=task_id)

    assert b"957694667" in job_artifacts.dirac_log
    assert job_artifacts.workflow.find(".//StepInstance/Parameter[@name='applicationVersion']/value").text == "v46r8"

    assert len(job_artifacts.steps) == 1

    assert b"DaVinciInitAlg    SUCCESS" in job_artifacts.steps[0].application_log
    assert job_artifacts.steps[0].prodrun_config["input"]["files"] == [
        "LFN:/lhcb/MC/2018/ALLSTREAMS.MDST/00097909/0000/00097909_00000169_7.AllStreams.mdst"
    ]
    assert job_artifacts.steps[0].summary_xml.findall(".//input/file")[0].attrib["status"] == "part"



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
""" Analyse XMLSummary test module."""
# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long
import pytest
from pathlib import Path
from textwrap import dedent

from DIRAC import S_OK, S_ERROR
from DIRAC.TransformationSystem.Client.FileReport import FileReport
from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary

# mocks
from LHCbDIRAC.Workflow.Modules.test.mock_Commons import (
    step_commons,
    wf_commons,
    prod_id,
    prod_job_id,
    wms_job_id,
    workflowStatus,
    step_id,
    step_number,
)
from DIRAC.DataManagementSystem.Client.test.mock_DM import dm_mock
from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import bkc_mock
from LHCbDIRAC.Workflow.Modules.AnalyseXMLSummary import AnalyseXMLSummary


# Helper functions
XML_SUMMARY_FILE = "XMLSummary.xml"


def prepare_XMLSummary_file(xml_summary, content):
    with open(xml_summary, "w") as f:
        f.write(content)
    return XMLSummary(xml_summary)


@pytest.fixture
def axlf(mocker):
    """Fixture for AnalyseXMLSummary module."""
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator")
    mocker.patch("LHCbDIRAC.Workflow.Modules.AnalyseXMLSummary.AnalyseXMLSummary._resolveInputVariables")
    mocker.patch("LHCbDIRAC.Workflow.Modules.AnalyseXMLSummary.NotificationClient.sendMail", return_value=S_OK())

    axlf = AnalyseXMLSummary(bkClient=bkc_mock, dm=dm_mock)
    mocker.patch.object(axlf, "jobReport", autospec=True)
    axlf.jobReport.setApplicationStatus.return_value = S_OK()
    axlf.fileReport = FileReport()
    axlf.XMLSummary = XML_SUMMARY_FILE

    yield axlf

    # Teardown
    Path(XML_SUMMARY_FILE).unlink()


# Test scenarios
def test_analyseXMLSummary_basic_success(axlf):
    """Test basic success scenario."""
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="full">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "True"
    assert axlf.XMLSummary_o.step == "finalize"
    assert axlf.XMLSummary_o._outputsOK()
    assert not axlf.XMLSummary_o.inputFileStats["mult"]
    assert not axlf.XMLSummary_o.inputFileStats["other"]
    assert axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_called_once()
    axlf.nc.sendMail.assert_not_called()
    assert axlf.fileReport.statusDict == {}


def test_analyseXMLSummary_previousError_success(axlf):
    """Test success scenario with previous error: stepStatus = S_ERROR()."""
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="full">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "True"
    assert axlf.XMLSummary_o.step == "finalize"
    assert axlf.XMLSummary_o._outputsOK()
    assert not axlf.XMLSummary_o.inputFileStats["mult"]
    assert not axlf.XMLSummary_o.inputFileStats["other"]
    assert axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_ERROR(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_not_called()
    axlf.nc.sendMail.assert_not_called()
    assert axlf.fileReport.statusDict == {}


def test_analyseXMLSummary_badInput_success(axlf):
    """Test success scenario with part and fail input not part of the input data list."""
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="part">200</file>
                        <file GUID="CCE96809-4FC6-F623-61F5-003048F35253" name="LFN:00012478_00000533_1.sim"
                        status="fail">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "True"
    assert axlf.XMLSummary_o.step == "finalize"
    assert axlf.XMLSummary_o._outputsOK()
    assert not axlf.XMLSummary_o.inputFileStats["mult"]
    assert not axlf.XMLSummary_o.inputFileStats["other"]
    assert axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_called_once()
    axlf.nc.sendMail.assert_not_called()
    assert axlf.fileReport.statusDict == {}


def test_analyseXMLSummary_partInput_success(axlf):
    """Test success scenario with part input part of the input data list."""
    # Input is 'part' and is part of the input data list but the number of events is not -1
    axlf.inputDataList = ["00012478_00000532_1.sim"]
    axlf.numberOfEvents = 1

    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="part">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "True"
    assert axlf.XMLSummary_o.step == "finalize"
    assert axlf.XMLSummary_o._outputsOK()
    assert not axlf.XMLSummary_o.inputFileStats["mult"]
    assert not axlf.XMLSummary_o.inputFileStats["other"]
    assert axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_called_once()
    axlf.nc.sendMail.assert_not_called()
    assert axlf.fileReport.statusDict == {}


def test_analyseXMLSummary_notSuccess_fail(axlf):
    """Test failure scenario with success=False."""
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>False</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="part">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "False"
    assert axlf.XMLSummary_o.step == "finalize"
    assert axlf.XMLSummary_o._outputsOK()
    assert not axlf.XMLSummary_o.inputFileStats["mult"]
    assert not axlf.XMLSummary_o.inputFileStats["other"]
    assert not axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_called_once()
    assert axlf.fileReport.statusDict == {}


def test_analyseXMLSummary_badStep_fail(axlf):
    """Test failure scenario with step != finalize."""
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>execute</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="part">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "True"
    assert axlf.XMLSummary_o.step == "execute"
    assert axlf.XMLSummary_o._outputsOK()
    assert not axlf.XMLSummary_o.inputFileStats["mult"]
    assert not axlf.XMLSummary_o.inputFileStats["other"]
    assert not axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_called_once()
    assert axlf.fileReport.statusDict == {}


def test_analyseXMLSummary_badOutput_fail(axlf):
    """Test failure scenario with output status != full."""
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="part">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="fail">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "True"
    assert axlf.XMLSummary_o.step == "finalize"
    assert not axlf.XMLSummary_o._outputsOK()
    assert not axlf.XMLSummary_o.inputFileStats["mult"]
    assert not axlf.XMLSummary_o.inputFileStats["other"]
    assert not axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_called_once()
    assert axlf.fileReport.statusDict == {}


def test_analyseXMLSummary_badInput_fail(axlf):
    """Test failure scenario with input status = mult."""
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="mult">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "True"
    assert axlf.XMLSummary_o.step == "finalize"
    assert axlf.XMLSummary_o._outputsOK()
    assert axlf.XMLSummary_o.inputFileStats["mult"]
    assert not axlf.XMLSummary_o.inputFileStats["other"]
    assert not axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_called_once()
    assert axlf.fileReport.statusDict == {}


def test_analyseXMLSummary_badInput2_fail(axlf):
    """Test failure scenario with an unknown input status (weoweo)."""
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="weoweo">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "True"
    assert axlf.XMLSummary_o.step == "finalize"
    assert axlf.XMLSummary_o._outputsOK()
    assert not axlf.XMLSummary_o.inputFileStats["mult"]
    assert axlf.XMLSummary_o.inputFileStats["other"]
    assert not axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_called_once()
    assert axlf.fileReport.statusDict == {}


def test_analyseXMLSummary_badInput3_fail(axlf):
    """Test failure scenario with input status = fail."""
    axlf.inputDataList = ["00012478_00000532_1.sim"]

    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="fail">200</file>
                        <file GUID="CCE96709-5BE9-E012-41BD-004048E36253" name="LFN:00012478_00000533_1.sim"
                        status="fail">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "True"
    assert axlf.XMLSummary_o.step == "finalize"
    assert axlf.XMLSummary_o._outputsOK()
    assert not axlf.XMLSummary_o.inputFileStats["mult"]
    assert not axlf.XMLSummary_o.inputFileStats["other"]
    assert not axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_called_once()
    assert axlf.fileReport.statusDict == {"00012478_00000532_1.sim": "Problematic"}


def test_analyseXMLSummary_badInput4_fail(axlf):
    """Test failure scenario with input status = part."""
    # Input is 'part' and is part of the input data list but the number of events is -1 (by default)
    axlf.inputDataList = ["00012478_00000532_1.sim"]

    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:00012478_00000532_1.sim"
                        status="part">200</file>
                        <file GUID="CCE96709-5BE9-E012-41BD-004048E36253" name="LFN:00012478_00000533_1.sim"
                        status="part">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )
    axlf.XMLSummary_o = prepare_XMLSummary_file(axlf.XMLSummary, xml_content)

    assert axlf.XMLSummary_o.success == "True"
    assert axlf.XMLSummary_o.step == "finalize"
    assert axlf.XMLSummary_o._outputsOK()
    assert not axlf.XMLSummary_o.inputFileStats["mult"]
    assert not axlf.XMLSummary_o.inputFileStats["other"]
    assert not axlf.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"]
    axlf.jobReport.setApplicationStatus.assert_called_once()
    assert axlf.fileReport.statusDict == {"00012478_00000532_1.sim": "Problematic"}



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Dummy module to be removed as soon
as all the Transformations with it in the workflow XML are gone
"""
import os

from DIRAC import gLogger
from DIRAC.Core.Utilities.ReturnValues import S_OK

from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class ErrorLogging(ModuleBase):
    def __init__(self):
        """c'tor."""

        self.log = gLogger.getSubLogger("ErrorLogging")
        super().__init__(self.log)

    def execute(self, *args, **kwargs):
        self.log.info(f"===== Executing fake ErrorLogging module ===== ")
        return S_OK()



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Defines patterns for errors that can be found in LHCb jobs.

NOTE: The order of classes is important for defining the priorities between them.
All reason classes are now defined in reasons.yaml and loaded dynamically.
Classes with custom behavior use mixin classes defined in this module.

The order in reasons.yaml determines the priority of pattern matching.
"""
from __future__ import annotations

import re

from .base_reasons import KnownUnrecoverableFailure, ReasonMixin
from .reason_models import ReasonsConfig

RE_MEMORY_ADDRESS = re.compile(r"(0x[0-9a-f]{4,})(?=[^\w])")


# Mixin classes for customizing failure reason behavior
class FatalGaudiErrorMixin(ReasonMixin):
    """Mixin for FatalGaudiError to extract full error context."""

    @property
    def match_string(self) -> str:
        """Extract FATAL/ERROR lines leading up to the match."""
        log = self._job.files.steps[-1].application_log
        old_start = log.rfind(b"\n", 0, self._start)
        maybe_skip = None
        while (start := log.rfind(b"\n", 0, old_start)) > 0:
            line = log[start:old_start]
            if b"FATAL" in line or b"ERROR" in line:
                old_start = start
                maybe_skip = None
            elif maybe_skip is None:
                maybe_skip = old_start
                old_start = start
            else:
                start = maybe_skip
                break
        else:
            start = 0

        return log[start : self._end].decode(errors="backslashreplace")


class SegfaultTracebackMixin(ReasonMixin):
    """Mixin for UnknownSegmentationFaultWithTraceback to filter and anonymize traceback."""

    @property
    def match_string(self) -> str:
        """Extract traceback with memory addresses masked and Python internals filtered."""
        log = self._job.files.steps[-1].application_log
        match_string = log[self._start : self._end].decode(errors="backslashreplace")

        # Replace anything which looks like a memory address
        match_string = RE_MEMORY_ADDRESS.sub(
            lambda match: "0x" + "X" * len(match.group(0)),
            match_string,
        )
        # Try to remove superfluous parts of the match
        raw_lines = match_string.splitlines(keepends=True)
        filtered_lines = None
        for line in raw_lines:
            if filtered_lines is None:
                # Remove everything before the first line of ====
                if line.startswith("===="):
                    filtered_lines = []
            elif line.startswith("====") or "in _Py" in line or "in Py" in line:
                # Only keep parts of the traceback which are LHCb specific
                # If we reach the second line of ==== we're also done
                break
            else:
                filtered_lines.append(line)
        return "".join(filtered_lines) if filtered_lines else match_string


# Special case: KnownCorruptedFile has empty patterns, needs to be defined manually
class KnownCorruptedFile(KnownUnrecoverableFailure):
    # This is a dummy reason that can only be manually initialised
    patterns = []
    explanation = "This file is known contain some corrupted data."


def _load_reasons_from_yaml() -> dict[str, type]:
    """Load and instantiate all reason classes defined in YAML.

    This function loads the reasons.yaml configuration file and dynamically
    creates Python classes for each reason definition. Classes with mixins
    are automatically composed with the mixin class in the MRO.

    Returns:
        dict: Mapping of reason class names to their class objects
    """
    try:
        config = ReasonsConfig.load_from_yaml()

        # Create all classes from YAML (mixins are handled in to_class())
        all_classes = {}
        for reason_def in config.reasons:
            reason_class = reason_def.to_class()
            all_classes[reason_def.name] = reason_class

        return all_classes

    except Exception as e:
        # If YAML loading fails, provide a helpful error message
        import sys

        print(f"ERROR: Failed to load reasons from YAML: {e}", file=sys.stderr)
        print("Make sure reasons.yaml exists and is valid.", file=sys.stderr)
        raise


# Load all YAML-defined reason classes and add them to the module namespace
# This makes them available for import like: from .reasons import FunctorCompilationFailure
_yaml_reason_classes = _load_reasons_from_yaml()
globals().update(_yaml_reason_classes)

# Export all reason class names and mixins for `from .reasons import *`
__all__ = [
    "RE_MEMORY_ADDRESS",
    "KnownCorruptedFile",
    "FatalGaudiErrorMixin",
    "SegfaultTracebackMixin",
] + list(_yaml_reason_classes.keys())



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
# from __future__ import annotations

__all__ = [
    "analyze_prod",
    "summarize_prod",
    "group_by_problem",
    "file_exists",
]

import asyncio
import importlib.resources
import json
import logging
import os
import re
import sys
from collections import defaultdict, Counter
from itertools import chain
from operator import attrgetter, itemgetter
from types import TracebackType
from urllib.parse import urlencode

import Levenshtein
from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient
from pathlib import Path
from pydantic import TypeAdapter
from rich import print

from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from .file_access import read_zip, file_exists
from .base_reasons import LogFileMatch, RecoverableFailure, KnownUnrecoverableFailure, EventTimeout, UnknownFailure
from .job_analyzer import (
    Job,
    LFNDebugInfo,
    PreAnalysisResult,
    PostAnalysisResult,
    LogFileMatches,
    LogFileMatch,
    AggregatedLFNDebugInfo,
    ReasonInfo,
    LbCondaJobsUnsupported,
)
from .lbconda import analyse_log as analyse_lbconda_log
from .utils import presorted_groupby, unsorted_groupby, ProgressUpdater, old_to_new_log_url, call_dirac


RE_LOG_SE = re.compile(r'href="(https://[^"/]+(?:/lhcb-dirac-logse/)?(/[^"]+)/)"')
LEVENSHTEIN_RATIO_THRESHOLD = 0.85
GroupedProblems = dict[str, dict[str, list[AggregatedLFNDebugInfo]]]
CacheType = dict[int, dict[str, LFNDebugInfo]]

logger = logging.getLogger(__name__)


def key_prioritize(reason: LFNDebugInfo) -> tuple[int, int, int, int, int, int]:
    """Key function to sort LFNDebugInfo such that the most important ones are first."""
    if reason.maybe_reason is None:
        key = reason.n_input_files, sys.maxsize, sys.maxsize, -reason.job_id, -sys.maxsize
    else:
        reason_type = LogFileMatch.name_to_reason[reason.maybe_reason.type]  # pylint: disable=unsubscriptable-object
        for i, base_type in enumerate([KnownUnrecoverableFailure, UnknownFailure, (RecoverableFailure, EventTimeout)]):
            if issubclass(reason_type, base_type):
                key = (
                    i,
                    LogFileMatch.name_to_id[reason.maybe_reason.type],  # pylint: disable=unsubscriptable-object
                    -reason.job_id,
                    -len(reason.maybe_reason.match),
                )
                break
        else:
            raise NotImplementedError(reason.maybe_reason.type)

    match reason.pre_analysis:
        case PreAnalysisResult.NO_EVENTS_SELECTED:
            return reason.n_input_files, -1, *key
        case PreAnalysisResult.MAYBE_POST_APPLICATION_FAILURE:
            return reason.n_input_files, 0, *key
        case PreAnalysisResult.PROBABLY_OK:
            return reason.n_input_files, 1, *key
        case PreAnalysisResult.INPUT_DATA_ALREADY_PROCESSED:
            return reason.n_input_files, 2, *key
        case PreAnalysisResult.NEEDS_LOG_ANALYSIS:
            return reason.n_input_files, 3, *key
        case PreAnalysisResult.UNKNOWN_ERROR_ON_JOB_SUCCESS:
            return reason.n_input_files, 4, *key
        case PreAnalysisResult.NEEDS_GROUP_SIZE_REDUCTION:
            return reason.n_input_files, 5, *key
        case PreAnalysisResult.NOT_IN_SUMMARY_XML:
            return reason.n_input_files, 6, *key
        case PreAnalysisResult.NO_XML_SUMMARY:
            return reason.n_input_files, 7, *key
        case PreAnalysisResult.STALLED:
            return reason.n_input_files, 8, *key
        case PreAnalysisResult.FAILED_INPUT_DATA_RESOLUTION:
            return reason.n_input_files, 9, *key
        case PreAnalysisResult.NO_LOG_FILE:
            return reason.n_input_files, 10, *key
        case PreAnalysisResult.NO_LOGSE_JOB_PARAMETER:
            return reason.n_input_files, 11, *key
    raise NotImplementedError(reason)


async def analyze_prod(
    tid: int,
    *,
    progress: ProgressUpdater = ProgressUpdater(),
    lfns: list[str] | None = None,
    statuses: list[str] = ["MaxReset", "Problematic"],
) -> list[LFNDebugInfo]:
    print(f"{tid}: Getting files")
    files = await call_dirac(
        TransformationClient().getTransformationFiles,
        {"TransformationID": tid} | ({"LFN": lfns} if lfns else {"Status": statuses}),
        columns=["FileID", "LFN"],
    )
    file_ids = {f["FileID"]: f"LFN:{f['LFN']}" for f in files}
    if len(file_ids) == 0:
        print(f"{tid}: No files found, returning early")
        return []

    print(f"{tid}: Getting tasks for {len(file_ids)} files")
    res = await call_dirac(
        TransformationClient().getTableDistinctAttributeValues,
        "TransformationFileTasks",
        ["TaskID"],
        {"TransformationID": tid, "FileID": list(file_ids)},
    )
    tasks = await call_dirac(
        TransformationClient().getTransformationTasks,
        {"TransformationID": tid, "TaskID": res["TaskID"]},
        inputVector=True,
    )
    task_id_to_input_lfns = {
        k: [f"LFN:{x}" for x in next(v)["InputVector"].split(";")]
        for k, v in presorted_groupby(tasks, key=itemgetter("TaskID"))
    }
    job_id_to_task_info = {
        int(k): next(v) for k, v in presorted_groupby(tasks, key=itemgetter("ExternalID")) if int(k) != 0
    }
    with ProductionAnalysisCache(tid) as cache:
        results = await cache.get_results(tid, job_id_to_task_info, files, task_id_to_input_lfns, progress)
    return list(chain.from_iterable(results.values()))


class ProductionAnalysisCache:
    """Cache for production analysis results with automatic persistence."""

    def __init__(self, cache_key: str | int, *, cache_directory: Path | None = None) -> None:
        self._cache_key = str(cache_key)
        self._cache_directory = cache_directory or self._pick_cache_directory()
        self._path = self._cache_directory / f"{self._cache_key}.json"
        self._cache: CacheType = self._load_cache()
        self._modified = False

    @property
    def cache_key(self) -> str:
        """The cache key used for this cache instance."""
        return self._cache_key

    @property
    def path(self) -> Path:
        """The filesystem path where the cache is stored."""
        return self._path

    def _pick_cache_directory(self) -> Path:
        """Determine the appropriate cache directory following XDG standards."""
        cache_home = Path(os.environ.get("XDG_CACHE_HOME", Path.home() / ".cache"))
        cache_dir = cache_home / "lhcbdirac/log-analysis-cache"
        cache_dir.mkdir(parents=True, exist_ok=True)
        return cache_dir

    def _load_cache(self) -> CacheType:
        """Load cache from disk if it exists."""
        if not self._path.is_file():
            return {}

        try:
            return TypeAdapter(CacheType).validate_json(self._path.read_text())
        except (json.JSONDecodeError, ValueError) as e:
            logger.warning(f"Failed to load cache from {self._path}: {e}. Starting with empty cache.")
            return {}

    def save(self) -> None:
        """Persist the cache to disk."""
        if not self._modified:
            return

        try:
            self._path.parent.mkdir(parents=True, exist_ok=True)
            self._path.write_bytes(TypeAdapter(CacheType).dump_json(self._cache))
            self._modified = False
        except OSError as e:
            logger.error(f"Failed to save cache to {self._path}: {e}")

    def __enter__(self) -> "ProductionAnalysisCache":
        """Context manager entry."""
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Context manager exit - automatically save cache."""
        self.save()

    def __getitem__(self, job_id: int) -> dict[str, LFNDebugInfo]:
        return self._cache.setdefault(job_id, {})

    def __contains__(self, job_id: int) -> bool:
        return job_id in self._cache

    def __setitem__(self, job_id: int, value: dict[str, LFNDebugInfo]) -> None:
        self._cache[job_id] = value
        self._modified = True

    async def get_results(
        self,
        tid: int | str,
        job_id_to_task_info: dict[int, dict[str, str]],
        files: list[dict[str, str]],
        task_id_to_input_lfns: dict[int, list[str]],
        progress: ProgressUpdater,
    ) -> dict[str, list[LFNDebugInfo]]:
        """
        Get analysis results, using cache where available and analyzing new jobs.

        Args:
            tid: Transformation ID
            job_id_to_task_info: Mapping of job IDs to task information
            files: List of file information dictionaries
            task_id_to_input_lfns: Mapping of task IDs to input LFN lists
            progress: Progress updater for tracking analysis

        Returns:
            Dictionary mapping LFNs to lists of debug information
        """
        results = self.get_cached_results(files, job_id_to_task_info)
        if not job_id_to_task_info:
            return results

        new_debug_infos = await _analyze_jobs(tid, job_id_to_task_info, task_id_to_input_lfns, progress)
        for debug_info in new_debug_infos:
            self[debug_info.job_id][debug_info.lfn] = debug_info
            if debug_info.lfn in results:
                results[debug_info.lfn].append(debug_info)

        self.save()
        return results

    def get_cached_results(
        self,
        files: list[dict[str, str]],
        job_id_to_task_info: dict[int, dict[str, str]],
    ) -> dict[str, list[LFNDebugInfo]]:
        """
        Retrieve cached results and remove cached jobs from the work queue.

        Args:
            files: List of file information dictionaries
            job_id_to_task_info: Mapping of job IDs to task information (modified in-place)

        Returns:
            Dictionary mapping LFNs to lists of cached debug information
        """
        results: dict[str, list[LFNDebugInfo]] = {f"LFN:{x['LFN']}": [] for x in files}
        cached_job_ids = [job_id for job_id in job_id_to_task_info if job_id in self]

        for job_id in cached_job_ids:
            job_id_to_task_info.pop(job_id)
            for lfn, debug_info in self[job_id].items():
                if lfn in results:
                    results[lfn].append(debug_info)

        return results


async def guess_log_lfn_if_needed(job_id, parameters, job_id_to_task_info: dict[int, dict[str, str]]) -> None:
    if "Log URL" in parameters:
        return

    input_lfns = job_id_to_task_info[job_id]["InputVector"].split(";")
    if not all(x.startswith("/lhcb/") for x in input_lfns):
        raise NotImplementedError("Cannot guess log LFN", job_id, input_lfns)
    parts = input_lfns[0].split("/")[:4]
    parts.append("LOG")
    parts.append(str(job_id_to_task_info[job_id]["TransformationID"]).rjust(8, "0"))
    task_id = str(job_id_to_task_info[job_id]["TaskID"]).rjust(8, "0")
    parts.append(task_id[:4])
    parts.append(task_id)
    lfn = "/".join(parts)
    if await file_exists(f"{lfn}.zip"):
        print("Guessed log LFN for job", job_id, ":", lfn)
        parameters["Log URL"] = f'<a href="https://lhcb-dirac-logse.web.cern.ch{lfn}/">'


async def _analyze_jobs(tid, job_id_to_task_info, task_id_to_input_lfns, progress) -> list[LFNDebugInfo]:
    print(f"{tid}: Getting Job parameters for {len(job_id_to_task_info)} jobs")
    job_parameters = await call_dirac(JobMonitoringClient().getJobParameters, list(job_id_to_task_info), ["Log URL"])
    assert len(job_parameters) == len(job_id_to_task_info)
    job_id_to_status = await call_dirac(JobMonitoringClient().getJobsStates, list(job_id_to_task_info))

    async with asyncio.TaskGroup() as tg:
        for job_id, parameters in job_parameters.items():
            tg.create_task(guess_log_lfn_if_needed(job_id, parameters, job_id_to_task_info))

    results: list[LFNDebugInfo] = []
    log_urls = {}
    read_tasks = {}
    site_missing = defaultdict(list)
    for job_id, parameters in job_parameters.items():
        if not parameters:
            input_lfns = set(task_id_to_input_lfns[job_id_to_task_info[job_id]["TaskID"]])
            for lfn in input_lfns:
                if job_id_to_status[job_id]["MinorStatus"] == "Watchdog identified this job as stalled":
                    pre_analysis = PreAnalysisResult.STALLED
                elif "Failed Input Data Resolution" in job_id_to_status[job_id]["ApplicationStatus"]:
                    pre_analysis = PreAnalysisResult.FAILED_INPUT_DATA_RESOLUTION
                else:
                    pre_analysis = PreAnalysisResult.NO_LOGSE_JOB_PARAMETER
                debug_info = LFNDebugInfo(lfn=lfn, job_id=job_id, pre_analysis=pre_analysis, transform_id=tid)
                site_missing[job_id].append(debug_info)
                results.append(debug_info)
        elif match := RE_LOG_SE.search(parameters["Log URL"]):
            if "Input Data Already Processed" in job_id_to_status[job_id]["ApplicationStatus"]:
                pre_analysis = PreAnalysisResult.INPUT_DATA_ALREADY_PROCESSED
                input_lfns = set(task_id_to_input_lfns[job_id_to_task_info[job_id]["TaskID"]])
                for lfn in input_lfns:
                    debug_info = LFNDebugInfo(
                        lfn=lfn,
                        job_id=job_id,
                        pre_analysis=pre_analysis,
                        log_url=old_to_new_log_url(match.group(1)),
                        transform_id=tid,
                    )
                    site_missing[job_id].append(debug_info)
                    results.append(debug_info)
            else:
                log_urls[job_id] = old_to_new_log_url(match.group(1))
                # TODO: Add timeout for reading the zip file
                read_tasks[read_zip(job_id, f"{match.group(2)}.zip")] = job_id
        else:
            raise NotImplementedError(parameters)

    lb_conda_jobs = []
    progress.advance_total(len(read_tasks))
    for future in asyncio.as_completed(read_tasks):
        job_id, result = await future
        task_id = job_id_to_task_info[job_id]["TaskID"]
        # TODO: Use an excetion and log?
        if result:
            job = Job(tid, task_id, result, job_id=job_id, url=log_urls[job_id])
            try:
                for lfn, debug_info in job.problems_by_lfn.items():
                    results.append(debug_info)
            except LbCondaJobsUnsupported:
                lb_conda_jobs.append(job)
        progress.advance()

    for job in lb_conda_jobs:
        input_files = job.files.steps[0].prodrun_config["input"]["files"]
        if len(input_files) > 1:
            for input_file in input_files:
                results.append(
                    LFNDebugInfo(
                        lfn=input_file,
                        pre_analysis=PreAnalysisResult.NEEDS_GROUP_SIZE_REDUCTION,
                        log_url=job.url,
                        job_id=job.job_id,
                        site=job.site,
                        peak_memory=job.peak_memory,
                        transform_id=job.transform_id,
                    )
                )
            continue
        matches = analyse_lbconda_log(job.files.steps[-1].application_log)
        if len(matches) == 0:
            raise NotImplementedError(job)
        else:
            results.append(
                LFNDebugInfo(
                    lfn=input_files[0],
                    pre_analysis=PreAnalysisResult.NEEDS_LOG_ANALYSIS,
                    log_url=job.url,
                    job_id=job.job_id,
                    site=job.site,
                    peak_memory=job.peak_memory,
                    transform_id=job.transform_id,
                    maybe_reason=matches[0],
                )
            )
        progress.advance()

    input_data = await call_dirac(JobMonitoringClient().getInputData, [r.job_id for r in results])
    for debug_info in results:
        debug_info.n_input_files = len(input_data[debug_info.job_id])

    if site_missing:
        print(f"{tid}: Getting missing site info for {len(site_missing)} jobs")
        for job_id, info in (await call_dirac(JobMonitoringClient().getJobsSites, list(site_missing))).items():
            for debug_info in site_missing[job_id]:
                debug_info.site = info["Site"]

    return results


def summarize_prod(debug_infos: list[LFNDebugInfo]) -> list[AggregatedLFNDebugInfo]:
    """Apply corrections that we can only infer by looking at the transformation as a whole"""
    results = []
    for _, debug_infos in unsorted_groupby(debug_infos, key=attrgetter("lfn")):
        debug_infos = sorted(debug_infos, key=key_prioritize)
        debug_info = debug_infos[0]
        info = {"attempts": len(debug_infos), "occurrences": 0, "sites": Counter(), "log_urls": []}

        if debug_info.maybe_reason is None:
            for past_debug_info in debug_infos:
                if debug_info.pre_analysis == past_debug_info.pre_analysis:
                    info["sites"][past_debug_info.site] += 1
                    info["occurrences"] += 1
                    if past_debug_info.log_url:
                        info["log_urls"].append(past_debug_info.log_url)
        else:
            if debug_info.pre_analysis == PreAnalysisResult.NEEDS_GROUP_SIZE_REDUCTION:
                info["post_analysis"] = PostAnalysisResult.NEEDS_GROUP_SIZE_REDUCTION

            if debug_info.maybe_reason.type == "XRDOperationExpired":
                if debug_info.pre_analysis == PreAnalysisResult.STALLED:
                    info["post_analysis"] = PostAnalysisResult.NEEDS_DOWNLOAD_INPUT_DATA
                elif debug_info.maybe_reason and debug_info.maybe_reason.type in {
                    "XRDOperationExpired",
                    "EventWatchdog",
                }:
                    info["post_analysis"] = PostAnalysisResult.NEEDS_DOWNLOAD_INPUT_DATA

            if debug_info.maybe_reason.type == "MaybeCorruptedFile":
                module_name = "LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis"
                with importlib.resources.files(module_name).joinpath("known_corruption.json").open("r") as f:
                    known_corruption = json.load(f)
                if known_corruption.get(debug_info.lfn.split(":", 1)[1]):
                    explanation = (
                        "This file is known to contain some corrupt events. It "
                        "likely cannot be recovered for this production however "
                        "we typically don't delete these files as other analyses "
                        "might not depend on the bad branches."
                    )
                    debug_info.maybe_reason = ReasonInfo(
                        type="KnownCorruptedFile", match=debug_info.maybe_reason.match, explanation=explanation
                    )

            for past_debug_info in debug_infos:
                past_reason = past_debug_info.maybe_reason
                if past_reason is None or past_reason.type != debug_info.maybe_reason.type:
                    continue
                if Levenshtein.ratio(past_reason.match, debug_info.maybe_reason.match) > LEVENSHTEIN_RATIO_THRESHOLD:
                    info["occurrences"] += 1
                    info["sites"][past_debug_info.site] += 1
                    info["log_urls"].append(past_debug_info.log_url)

        if debug_info.pre_analysis != PreAnalysisResult.NOT_IN_SUMMARY_XML and debug_info.peak_memory:
            # Reading corrupt data can trigger excessive memory consumption and shouldn't be considered
            if not (
                debug_info.maybe_reason
                and debug_info.maybe_reason.type
                in {"XRDOperationExpired", "MaybeCorruptedFile", "KnownCorruptedFile", "XRDIssue"}
            ):
                if debug_info.peak_memory > 10 * 1024**3:
                    info["post_analysis"] = PostAnalysisResult.EXTREMELY_HIGH_MEMORY_USAGE
                elif debug_info.peak_memory > 5 * 1024**3 and debug_info.maybe_reason is None:
                    info["post_analysis"] = PostAnalysisResult.HIGH_MEMORY_USAGE

        if (
            debug_info.peak_memory
            and debug_info.site in {"LCG.IN2P3.fr"}
            and getattr(debug_info.maybe_reason, "type", None) in {"MaybeOOMKiller", "FunctorCompilationFailure"}
        ):
            # Some sites have started enforcing memory limits via cgroups and will need to be sent to the hospital
            if debug_info.peak_memory > 2.8 * 1024**3:
                info["post_analysis"] = PostAnalysisResult.TOO_MUCH_MEMORY_FOR_CGROUPS

        if len(debug_infos) > 20:
            info["post_analysis"] = PostAnalysisResult.TOO_MANY_ATTEMPTS

        results.append(AggregatedLFNDebugInfo.model_validate(debug_info.model_dump(mode="json") | info))

    return results


def group_by_problem(all_problems: list[LogFileMatches]) -> GroupedProblems:
    def key_func(x: LogFileMatch) -> tuple[str, str, str]:
        if x.post_analysis:
            return (x.post_analysis, "", "")
        if x.maybe_reason:
            return ("", x.maybe_reason.type, "")
        return ("", "", x.pre_analysis)

    summarized = {}
    for (post_analysis, reason_type, pre_analysis), matches in unsorted_groupby(all_problems, key=key_func):
        if sum(map(bool, (post_analysis, reason_type, pre_analysis))) != 1:
            raise NotImplementedError("post_analysis/reason_type/pre_analysis are mutually exclusive when grouping")

        if reason_type:
            assert reason_type not in summarized
            summarized[reason_type] = {}
            for match_str, instances in unsorted_groupby(matches, key=attrgetter("maybe_reason.match")):
                for ref_match_str in summarized[reason_type]:
                    if len(match_str) > 100_000 or len(ref_match_str) > 100_000:
                        # Deduplicate lines in very large strings to avoid performance issues
                        a = "\n".join(sorted(set(match_str.splitlines())))
                        b = "\n".join(sorted(set(ref_match_str.splitlines())))
                        print(
                            f"Comparing large strings of lengths {len(match_str)} and {len(ref_match_str)} "
                            f"reduced to {len(a)} and {len(b)}"
                        )
                        ratio = Levenshtein.ratio(a, b)
                    else:
                        ratio = Levenshtein.ratio(match_str, ref_match_str)

                    if ratio > LEVENSHTEIN_RATIO_THRESHOLD:
                        summarized[reason_type][ref_match_str].extend(instances)
                        break
                else:
                    summarized[reason_type][match_str] = list(instances)
        else:
            summarized.setdefault("", {})[post_analysis or pre_analysis] = list(matches)
    return dict(summarized)



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

__all__ = ["presorted_groupby", "unsorted_groupby", "call_dirac", "old_to_new_log_url", "ProgressUpdater", "Progress"]

import asyncio
from dataclasses import dataclass
from functools import partial
from itertools import groupby as _groupby
from urllib.parse import urlencode, urlparse

from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise
from rich.progress import Progress as _Progress, MofNCompleteColumn, TimeElapsedColumn, TaskID as ProgressBarTaskID


@dataclass
class ProgressUpdater:
    progress: _Progress | None = None
    task: ProgressBarTaskID | None = None

    def advance(self):
        if self.progress is None or self.task is None:
            return
        self.progress.update(self.task, advance=1)

    def advance_total(self, n: int):
        if self.progress is None or self.task is None:
            return
        total = self.progress.tasks[self.progress.task_ids.index(self.task)].total
        total = total + n if total else n
        self.progress.update(self.task, total=total)


Progress = partial(_Progress, MofNCompleteColumn(), *_Progress.get_default_columns(), TimeElapsedColumn())


def presorted_groupby(iterable, key=None):
    return _groupby(iterable, key=key)


def unsorted_groupby(iterable, key=None):
    return _groupby(sorted(iterable, key=key), key=key)


async def call_dirac(func, *args, **kwargs):
    loop = asyncio.get_event_loop()
    result = await loop.run_in_executor(None, partial(func, *args, **kwargs))
    return returnValueOrRaise(result)


def old_to_new_log_url(url):
    if not url or "lhcb-productions.web.cern.ch" in url:
        return url
    path = urlparse(url).path.split("/")
    lfn = "/".join(path[:7])
    task_id = path[7]
    query = urlencode({"lfn": lfn, "task_name": task_id})
    return f"https://lhcb-productions.web.cern.ch/logs/?{query}"



###############################################################################
# (c) Copyright 2025 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from DIRAC import S_OK
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
from DIRAC.ConfigurationSystem.Client.Helpers import getVO
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getUsersInVO
from DIRAC.ConfigurationSystem.Agent.VOMS2CSAgent import VOMS2CSAgent as VOMS2CSAgentBase
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise

from LHCbDIRAC.ConfigurationSystem.Utilities.FenceIntegration import FenceIntegration


class VOMS2CSAgent(VOMS2CSAgentBase):
    """
    Extended VOMS2CSAgent to synchronise information with LBFence.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.csapi = CSAPI()
        self.lbfence = FenceIntegration()

    def initialize(self):
        """
        Extended initialize method.
        """
        init = super().initialize()
        if not init["OK"]:
            return init

        return init

    def execute(self):
        """
        Extended execute method to include Fence synchronization.
        """
        exec_res = super().execute()
        if not exec_res["OK"]:
            return exec_res

        vo = getVO("lhcb")

        ccid_to_end_date = {
            # We fall back to CCID if personId is not available
            # and have fallbacks for endDateInLHCbString because it's not clear how fence decides to return things...
            # however they are equivalent
            int(member.get("personId", member.get("CCID"))): member.get(
                "endDateInLHCbString", member.get("END_DATE_IN_LHCB_STRING")
            )
            for member in self.lbfence.get_all_memberships()
        }

        # Get all users from CS and update their AffiliationEnds for lhcb VO

        returnValueOrRaise(self.csapi.downloadCSData())
        users = returnValueOrRaise(self.csapi.describeUsers(getUsersInVO(vo)))
        cs_modified = False

        for nick, user in users.items():
            if not (person_id := user.get("CERNPersonId", None)):
                continue  # No Person ID, cannot map
            if end_date := ccid_to_end_date.get(int(person_id), None):
                affiliation_ends_current = user.get("AffiliationEnds", {})
                if affiliation_ends_current.get(vo, None) != end_date:
                    self.csapi.modifyUser(nick, {"AffiliationEnds": {**affiliation_ends_current, vo: end_date}})
                    self.log.info(f"Updated AffiliationEnds for {nick=} ({person_id=}) to {end_date=}.")
                    cs_modified = True

        if cs_modified:
            self.log.info("Committing AffiliationEnds changes to CS.")
            # Commit changes to CS
            returnValueOrRaise(self.csapi.commitChanges())
        else:
            self.log.info("No AffiliationEnds changes to commit to CS.")
        return S_OK()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test class for BookkeepingReport."""

import os
import time
import pytest
import xml.etree.ElementTree as ET

from pathlib import Path
from textwrap import dedent

import DIRAC
import LHCbDIRAC

from DIRAC import S_OK, S_ERROR
from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary
from LHCbDIRAC.Workflow.Modules.BookkeepingReport import BookkeepingReport
from LHCbDIRAC.Workflow.Modules.test.mock_Commons import (
    prod_id,
    prod_job_id,
    wms_job_id,
    workflowStatus,
    step_id,
    step_number,
    step_commons,
    wf_commons,
)


# Helper Functions
XML_SUMMARY_FILE = "XMLSummary.xml"
POOL_CATALOG_FILE = "pool_xml_catalog.xml"


def prepare_XMLSummary_file(xml_summary, content):
    with open(xml_summary, "w") as f:
        f.write(content)
    return XMLSummary(xml_summary)


def get_typed_parameter_value(name, root):
    """Find the value of a specific TypedParameter by its name."""
    for child in root:
        if child.tag == "TypedParameter" and child.attrib["Name"] == name:
            return child.attrib["Value"]
    return None


def get_output_file_details(output_file):
    """Extract details from an OutputFile element."""
    details = {
        "Name": output_file.attrib["Name"],
        "TypeName": output_file.attrib["TypeName"],
        "Parameters": {},
        "Replicas": [],
    }

    for elem in output_file:
        if elem.tag == "Parameter":
            details["Parameters"][elem.attrib["Name"]] = elem.attrib["Value"]
        elif elem.tag == "Replica":
            details["Replicas"].append({"Name": elem.attrib["Name"], "Location": elem.attrib["Location"]})

    return details


@pytest.fixture
def bkreport(mocker):
    """Fixture for BookkeepingReport module."""
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.RequestValidator")
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.ModuleBase._resolveInputVariables")
    mocker.patch("LHCbDIRAC.Workflow.Modules.ModuleBase.ModuleBase._resolveInputStep")

    bkreport = BookkeepingReport()
    mocker.patch.object(bkreport, "jobReport", autospec=True)
    bkreport.jobReport.setApplicationStatus.return_value = S_OK()

    step_commons[0]["listoutput"] = []
    step_commons[0]["inputData"] = []
    wf_commons[0].pop("BookkeepingLFNs", None)
    wf_commons[0].pop("LogFilePath", None)
    wf_commons[0].pop("ProductionOutputData", None)

    yield bkreport

    # Teardown
    xml_filename = f"bookkeeping_{step_id}.xml"
    Path(xml_filename).unlink(missing_ok=True)
    Path(XML_SUMMARY_FILE).unlink(missing_ok=True)
    Path(POOL_CATALOG_FILE).unlink(missing_ok=True)
    if bkreport.applicationLog:
        Path(bkreport.applicationLog).unlink(missing_ok=True)
    for output in step_commons[0]["listoutput"]:
        Path(output["outputDataName"]).unlink(missing_ok=True)


# Test Scenarios
def test_bkreport_prod_mcsimulation_success(bkreport):
    """Test successful execution of BookkeepingReport module."""
    # Mock the BookkeepingReport module
    bkreport.applicationName = "Gauss"
    bkreport.applicationVersion = "v49r10"
    bkreport.jobType = "MCSimulation"

    # This was obtained from a previous module (likely GaudiApplication)
    wf_commons[0]["BookkeepingLFNs"] = ["/lhcb/LHCb/Collision16/SIM/00209455/0000/00209455_00001537_1.sim"]
    wf_commons[0]["LogFilePath"] = "/lhcb/LHCb/Collision16/LOG/00209455/0000/"
    wf_commons[0]["ProductionOutputData"] = ["/lhcb/LHCb/Collision16/SIM/00209455/0000/00209455_00001537_1.sim"]

    step_commons[0]["StartTime"] = time.time() - 1000

    # Add input and output files
    # Input data should be None as we use Gauss (MCSimulation)
    step_commons[0]["inputData"] = None
    step_commons[0]["listoutput"] = [
        {"outputDataName": "00209455_00001537_1.sim", "outputDataType": "sim"},
    ]
    bkreport.applicationLog = "application.log"
    Path(bkreport.applicationLog).touch()
    Path(step_commons[0]["listoutput"][0]["outputDataName"]).touch()

    # Mock the XMLSummary object
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="1.0"
        xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd">
            <success>True</success>
            <step>finalize</step>
            <usage>
                <stat useOf="MemoryMaximum" unit="KB">2129228.0</stat>
            </usage>
            <input />
            <output>
                <file GUID="F2A331E0-C977-11EE-8689-D85ED3091B7C" name="PFN:00209455_00001537_1.sim" status="full">1</file>
            </output>
            <counters>
                <counter name="ConversionFilter/event with gamma conversion from">1</counter>
                <counter name="GaussGeo.Hcal/#energy">77</counter>
                <counter name="GaussGeo.Hcal/#hits">2644</counter>
                <counter name="GaussGeo.Hcal/#subhits">6262</counter>
                <counter name="GaussGeo.Hcal/#tslots">8391</counter>
                <counter name="GaussGeo.Ecal/#energy">963</counter>
                <counter name="GaussGeo.Ecal/#hits">18139</counter>
                <counter name="GaussGeo.Ecal/#subhits">45169</counter>
                <counter name="GaussGeo.Ecal/#tslots">52237</counter>
                <counter name="CounterSummarySvc/handled">79</counter>
            </counters>
            <lumiCounters />
        </summary>
        """
    )

    step_commons[0]["XMLSummary_o"] = prepare_XMLSummary_file(XML_SUMMARY_FILE, xml_content)

    # Pool XML catalog content
    pool_xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8" standalone="no" ?>
        <!-- Edited By POOL -->
        <!DOCTYPE POOLFILECATALOG SYSTEM "InMemory">
        <POOLFILECATALOG>
        <File ID="F2A331E0-C977-11EE-8689-D85ED3091B7C">
            <physical>
            <pfn filetype="ROOT" name="00209455_00001537_1.sim"/>
            </physical>
            <logical/>
        </File>
        </POOLFILECATALOG>
        """
    )
    with open(POOL_CATALOG_FILE, "w") as f:
        f.write(pool_xml_content)

    # Execute the module
    assert bkreport.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"], "Execution should succeed."

    # Check if the XML report file is created
    xml_filename = f"bookkeeping_{step_id}.xml"
    assert Path(xml_filename).exists(), "XML report file not created."

    # Validate the XML file
    tree = ET.parse(xml_filename)
    root = tree.getroot()

    # Extract fields from the XML and perform further operations
    assert root.tag == "Job", "Root tag should be Job."
    assert root.attrib["ConfigName"] == "aConfigName"
    assert root.attrib["ConfigVersion"] == "aConfigVersion"
    assert root.attrib["Date"] == bkreport.ldate
    assert root.attrib["Time"] == bkreport.ltime

    assert get_typed_parameter_value("ProgramName", root) == bkreport.applicationName
    assert get_typed_parameter_value("ProgramVersion", root) == bkreport.applicationVersion
    assert get_typed_parameter_value("DiracVersion", root) == LHCbDIRAC.__version__
    assert get_typed_parameter_value("Name", root) == step_id
    assert float(get_typed_parameter_value("ExecTime", root)) > 1000
    assert get_typed_parameter_value("CPUTIME", root) == "0"

    assert get_typed_parameter_value("FirstEventNumber", root) == "1"
    assert get_typed_parameter_value("StatisticsRequested", root) == str(bkreport.numberOfEvents)
    assert get_typed_parameter_value("NumberOfEvents", root) == str(bkreport.xf_o.outputEventsTotal)

    assert get_typed_parameter_value("Production", root) == prod_id
    assert get_typed_parameter_value("DiracJobId", root) == str(wms_job_id)
    assert get_typed_parameter_value("Location", root) == DIRAC.siteName()
    assert get_typed_parameter_value("JobStart", root) == f"{bkreport.ldatestart} {bkreport.ltimestart}"
    assert get_typed_parameter_value("JobEnd", root) == f"{bkreport.ldate} {bkreport.ltime}"
    assert get_typed_parameter_value("JobType", root) == "MCSimulation"

    assert get_typed_parameter_value("WorkerNode", root)
    assert get_typed_parameter_value("WNMEMORY", root)
    assert get_typed_parameter_value("WNCPUPOWER", root)
    assert get_typed_parameter_value("WNMODEL", root)
    assert get_typed_parameter_value("WNCACHE", root)
    assert get_typed_parameter_value("WNCPUHS06", root)
    assert get_typed_parameter_value("NumberOfProcessors", root) == str(bkreport.numberOfProcessors)

    # Input should be empty
    input_file = root.find("InputFile")
    assert input_file is None, "InputFile element should not be present."

    # Output should not be empty
    output_files = root.findall("OutputFile")
    assert output_files, "No OutputFile elements found."

    first_output_details = get_output_file_details(output_files[0])
    assert first_output_details["Name"] == wf_commons[0]["ProductionOutputData"][0]
    assert first_output_details["TypeName"] == "SIM"
    assert first_output_details["Parameters"]["FileSize"] == "0"
    assert "CreationDate" in first_output_details["Parameters"]
    assert "MD5Sum" in first_output_details["Parameters"]
    assert "Guid" in first_output_details["Parameters"]

    assert len(output_files) == 1


def test_bkreport_prod_mcsimulation_noinputoutput_success(bkreport):
    """Test successful execution of BookkeepingReport module:

    * No input files because bkreport.stepInputData is empty
    * No output files because bkreport.stepOutputData is empty
    * No pool xml catalog
    * Simulation conditions because the application used is Gauss
    """
    # Mock the BookkeepingReport module
    bkreport.applicationName = "Gauss"
    bkreport.applicationVersion = "v49r10"
    bkreport.jobType = "MCSimulation"

    # This was obtained from a previous module (likely GaudiApplication)
    wf_commons[0]["BookkeepingLFNs"] = ["/lhcb/LHCb/Collision16/SIM/00209455/0000/00209455_00001537_1"]
    wf_commons[0]["LogFilePath"] = "/lhcb/LHCb/Collision16/LOG/00209455/0000/"
    wf_commons[0]["ProductionOutputData"] = ["/lhcb/LHCb/Collision16/SIM/00209455/0000/00209455_00001537_1"]

    step_commons[0]["StartTime"] = time.time() - 1000

    # Mock the XMLSummary object
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" version="1.0"
        xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd">
            <success>True</success>
            <step>finalize</step>
            <usage>
                <stat useOf="MemoryMaximum" unit="KB">2129228.0</stat>
            </usage>
            <input />
            <output>
                <file GUID="F2A331E0-C977-11EE-8689-D85ED3091B7C" name="PFN:00211518_00024143_1.sim" status="full">1</file>
            </output>
            <counters>
                <counter name="ConversionFilter/event with gamma conversion from">1</counter>
                <counter name="GaussGeo.Hcal/#energy">77</counter>
                <counter name="GaussGeo.Hcal/#hits">2644</counter>
                <counter name="GaussGeo.Hcal/#subhits">6262</counter>
                <counter name="GaussGeo.Hcal/#tslots">8391</counter>
                <counter name="GaussGeo.Ecal/#energy">963</counter>
                <counter name="GaussGeo.Ecal/#hits">18139</counter>
                <counter name="GaussGeo.Ecal/#subhits">45169</counter>
                <counter name="GaussGeo.Ecal/#tslots">52237</counter>
                <counter name="CounterSummarySvc/handled">79</counter>
            </counters>
            <lumiCounters />
        </summary>
        """
    )

    step_commons[0]["XMLSummary_o"] = prepare_XMLSummary_file(XML_SUMMARY_FILE, xml_content)

    # Execute the module
    assert bkreport.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"], "Execution should succeed."

    # Check if the XML report file is created
    xml_filename = f"bookkeeping_{step_id}.xml"
    assert Path(xml_filename).exists(), "XML report file not created."

    # Validate the XML file
    tree = ET.parse(xml_filename)
    root = tree.getroot()

    # Extract fields from the XML and perform further operations
    assert root.tag == "Job", "Root tag should be Job."
    assert root.attrib["ConfigName"] == "aConfigName"
    assert root.attrib["ConfigVersion"] == "aConfigVersion"
    assert root.attrib["Date"] == bkreport.ldate
    assert root.attrib["Time"] == bkreport.ltime

    assert get_typed_parameter_value("ProgramName", root) == bkreport.applicationName
    assert get_typed_parameter_value("ProgramVersion", root) == bkreport.applicationVersion
    assert get_typed_parameter_value("DiracVersion", root) == LHCbDIRAC.__version__
    assert get_typed_parameter_value("Name", root) == step_id
    assert float(get_typed_parameter_value("ExecTime", root)) > 1000
    assert get_typed_parameter_value("CPUTIME", root) == "0"

    assert get_typed_parameter_value("FirstEventNumber", root) == "1"
    assert get_typed_parameter_value("StatisticsRequested", root) == str(bkreport.numberOfEvents)
    assert get_typed_parameter_value("NumberOfEvents", root) == str(bkreport.xf_o.outputEventsTotal)

    assert get_typed_parameter_value("Production", root) == prod_id
    assert get_typed_parameter_value("DiracJobId", root) == str(wms_job_id)
    assert get_typed_parameter_value("Location", root) == DIRAC.siteName()
    assert get_typed_parameter_value("JobStart", root) == f"{bkreport.ldatestart} {bkreport.ltimestart}"
    assert get_typed_parameter_value("JobEnd", root) == f"{bkreport.ldate} {bkreport.ltime}"
    assert get_typed_parameter_value("JobType", root) == "MCSimulation"

    assert get_typed_parameter_value("WorkerNode", root)
    assert get_typed_parameter_value("WNMEMORY", root)
    assert get_typed_parameter_value("WNCPUPOWER", root)
    assert get_typed_parameter_value("WNMODEL", root)
    assert get_typed_parameter_value("WNCACHE", root)
    assert get_typed_parameter_value("WNCPUHS06", root)
    assert get_typed_parameter_value("NumberOfProcessors", root) == str(bkreport.numberOfProcessors)

    # Input should be empty
    input_file = root.find("InputFile")
    assert input_file is None, "InputFile element should not be present."

    # Output should be empty
    output_file = root.find("OutputFile")
    assert output_file is None, "OutputFile element should not be present."


def test_bkreport_prod_mcreconstruction_success(bkreport):
    """Test successful execution of BookkeepingReport module."""
    # Mock the BookkeepingReport module
    bkreport.applicationName = "Boole"
    bkreport.applicationVersion = "v49r10"
    bkreport.jobType = "MCReconstruction"

    # This was obtained from a previous module (likely GaudiApplication)
    wf_commons[0]["BookkeepingLFNs"] = ["/lhcb/LHCb/Collision16/SIM/00209455/0000/00209455_00001537_1"]
    wf_commons[0]["LogFilePath"] = "/lhcb/LHCb/Collision16/LOG/00209455/0000/"
    wf_commons[0]["ProductionOutputData"] = ["/lhcb/LHCb/Collision16/SIM/00209455/0000/00209455_00001537_1"]

    step_commons[0]["StartTime"] = time.time() - 1000

    # Add input and output files
    step_commons[0]["inputData"] = "/lhcb/MC/2018/SIM/00212581/0000/00212581_00001446_1.sim"
    step_commons[0]["listoutput"] = [
        {"outputDataName": "00209455_00001537_1", "outputDataType": "digi"},
    ]
    bkreport.applicationLog = "application.log"
    Path(bkreport.applicationLog).touch()
    Path(step_commons[0]["listoutput"][0]["outputDataName"]).touch()

    # Mock the XMLSummary object
    xml_content = dedent(
        """<?xml version="1.0" encoding="UTF-8"?>
        <summary version="1.0" xsi:noNamespaceSchemaLocation="$XMLSUMMARYBASEROOT/xml/XMLSummary.xsd"
        xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                <success>True</success>
                <step>finalize</step>
                <usage>
                        <stat unit="KB" useOf="MemoryMaximum">866104.0</stat>
                </usage>
                <input>
                        <file GUID="CCE96707-4BE9-E011-81CD-003048F35252" name="LFN:0209455_00001537_1.sim"
                        status="full">200</file>
                </input>
                <output>
                        <file GUID="229BBEF1-66E9-E011-BBD0-003048F35252" name="PFN:00012478_00000532_2.xdigi"
                        status="full">200</file>
                </output>
        </summary>
        """
    )

    step_commons[0]["XMLSummary_o"] = prepare_XMLSummary_file(XML_SUMMARY_FILE, xml_content)

    # Execute the module
    assert bkreport.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        stepStatus=S_OK(),
    )["OK"], "Execution should succeed."

    # Check if the XML report file is created
    xml_filename = f"bookkeeping_{step_id}.xml"
    assert Path(xml_filename).exists(), "XML report file not created."

    # Validate the XML file
    tree = ET.parse(xml_filename)
    root = tree.getroot()

    # Extract fields from the XML and perform further operations
    assert root.tag == "Job", "Root tag should be Job."
    assert root.attrib["ConfigName"] == "aConfigName"
    assert root.attrib["ConfigVersion"] == "aConfigVersion"
    assert root.attrib["Date"] == bkreport.ldate
    assert root.attrib["Time"] == bkreport.ltime

    assert get_typed_parameter_value("ProgramName", root) == bkreport.applicationName
    assert get_typed_parameter_value("ProgramVersion", root) == bkreport.applicationVersion
    assert get_typed_parameter_value("DiracVersion", root) == LHCbDIRAC.__version__
    assert get_typed_parameter_value("Name", root) == step_id
    assert float(get_typed_parameter_value("ExecTime", root)) > 1000
    assert get_typed_parameter_value("CPUTIME", root) == "0"

    assert get_typed_parameter_value("FirstEventNumber", root) == "1"
    assert get_typed_parameter_value("StatisticsRequested", root) == str(bkreport.numberOfEvents)
    assert get_typed_parameter_value("NumberOfEvents", root) == str(bkreport.xf_o.inputEventsTotal)

    assert get_typed_parameter_value("Production", root) == prod_id
    assert get_typed_parameter_value("DiracJobId", root) == str(wms_job_id)
    assert get_typed_parameter_value("Location", root) == DIRAC.siteName()
    assert get_typed_parameter_value("JobStart", root) == f"{bkreport.ldatestart} {bkreport.ltimestart}"
    assert get_typed_parameter_value("JobEnd", root) == f"{bkreport.ldate} {bkreport.ltime}"
    assert get_typed_parameter_value("JobType", root) == "MCReconstruction"

    assert get_typed_parameter_value("WorkerNode", root)
    assert get_typed_parameter_value("WNMEMORY", root)
    assert get_typed_parameter_value("WNCPUPOWER", root)
    assert get_typed_parameter_value("WNMODEL", root)
    assert get_typed_parameter_value("WNCACHE", root)
    assert get_typed_parameter_value("WNCPUHS06", root)
    assert get_typed_parameter_value("NumberOfProcessors", root) == str(bkreport.numberOfProcessors)

    # Input should be empty
    input_file = root.find("InputFile")
    assert input_file is None, "InputFile element should not be present."

    # Output should not be empty
    output_files = root.findall("OutputFile")
    assert output_files, "No OutputFile elements found."

    first_output_details = get_output_file_details(output_files[0])
    assert first_output_details["Name"] == wf_commons[0]["ProductionOutputData"][0]
    assert first_output_details["TypeName"] == "DIGI"
    assert first_output_details["Parameters"]["FileSize"] == "0"
    assert "CreationDate" in first_output_details["Parameters"]
    assert "MD5Sum" in first_output_details["Parameters"]
    assert "Guid" in first_output_details["Parameters"]

    assert len(output_files) == 1

    # Because we are using Gauss, sim conditions should be present too
    simulation_condition = root.find("SimulationCondition")
    assert simulation_condition is None, "SimulationCondition element should not be present."


def test_bkreport_previousError_success(bkreport):
    """Test BookkeepingReport with a previous error."""
    # Mock the BookkeepingReport module
    bkreport.applicationName = "Gauss"
    bkreport.applicationVersion = "v49r10"
    bkreport.jobType = "MCSimulation"

    # Execute the module
    assert bkreport.execute(
        production_id=prod_id,
        prod_job_id=prod_job_id,
        wms_job_id=wms_job_id,
        workflowStatus=workflowStatus,
        wf_commons=wf_commons[0],
        step_commons=step_commons[0],
        step_number=step_number,
        step_id=step_id,
        # An error occurred in a previous step
        stepStatus=S_ERROR(),
    )["OK"], "Execution should succeed."

    # Check if the XML report file is created
    xml_filename = f"bookkeeping_{step_id}.xml"
    assert not Path(xml_filename).exists(), "XML report file was created."



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from __future__ import annotations

import argparse
import asyncio
import json
from pathlib import Path

from pydantic import TypeAdapter
import pytest

if __name__ == "__main__":
    import DIRAC

    DIRAC.initialize()

from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.job_analyzer import (
    LFNDebugInfo,
    AggregatedLFNDebugInfo,
)
from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.prod_analyzer import (
    GroupedProblems,
)
from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis import (
    analyze_prod,
    summarize_prod,
    group_by_problem,
)
from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis import output
from LHCbDIRAC.ProductionManagementSystem.Utilities.ProductionTools.LogAnalysis.test import deterministic_dump


ResultAdapter = TypeAdapter(list[LFNDebugInfo])
AggregatedResultAdapter = TypeAdapter(list[AggregatedLFNDebugInfo])
GroupedResultAdapter = TypeAdapter(GroupedProblems)
EXAMPLES_DIR = Path(__file__).parent / "examples" / "prods"
EXAMPLE_PROD_IDS = [int(x.name) for x in EXAMPLES_DIR.glob("*/")]


@pytest.mark.parametrize("prod_id", EXAMPLE_PROD_IDS)
def test_summarize_prod(prod_id: int):
    raw = ResultAdapter.validate_json((EXAMPLES_DIR / f"{prod_id}" / "raw.json").read_text())
    actual = AggregatedResultAdapter.dump_python(summarize_prod(raw), mode="json")
    # (EXAMPLES_DIR / f"{prod_id}" / "summarized.json").write_text(deterministic_dump(AggregatedResultAdapter, actual))
    expected = json.loads((EXAMPLES_DIR / f"{prod_id}" / "summarized.json").read_text())
    assert actual == expected


@pytest.mark.parametrize("prod_id", EXAMPLE_PROD_IDS)
def test_group_results(prod_id: int):
    summarized = AggregatedResultAdapter.validate_json((EXAMPLES_DIR / f"{prod_id}" / "summarized.json").read_text())
    actual = GroupedResultAdapter.dump_python(group_by_problem(summarized), mode="json")
    # (EXAMPLES_DIR / f"{prod_id}" / "grouped.json").write_text(deterministic_dump(GroupedResultAdapter, actual))
    expected = json.loads((EXAMPLES_DIR / f"{prod_id}" / "grouped.json").read_text())
    assert actual == expected


@pytest.mark.parametrize("prod_id", EXAMPLE_PROD_IDS)
def test_summarized_markdown(prod_id: int):
    grouped = GroupedResultAdapter.validate_json((EXAMPLES_DIR / f"{prod_id}" / "grouped.json").read_text())
    actual_markdown = output.to_markdown(grouped, is_summary=True)
    # (EXAMPLES_DIR / f"{prod_id}" / "summarized.md").write_text(actual_markdown)
    expected_markdown = (EXAMPLES_DIR / f"{prod_id}" / "summarized.md").read_text()
    assert actual_markdown == expected_markdown


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generate reference results of summarization")
    parser.add_argument("prod_id", type=int, help="Production ID")
    parser.add_argument("lfns", nargs="+", help="LFNs to analyze")
    args = parser.parse_args()

    prod_path = EXAMPLES_DIR / f"{args.prod_id}"
    raw_analysis = asyncio.run(analyze_prod(args.prod_id, lfns=args.lfns))
    summarized_analysis = summarize_prod(raw_analysis)
    grouped_analysis = group_by_problem(summarized_analysis)
    full_markdown = output.to_markdown(grouped_analysis, is_summary=False)
    summary_markdown = output.to_markdown(grouped_analysis, is_summary=True)

    prod_path.mkdir(exist_ok=False)
    (prod_path / "raw.json").write_text(deterministic_dump(ResultAdapter, raw_analysis))
    (prod_path / "summarized.json").write_text(deterministic_dump(AggregatedResultAdapter, summarized_analysis))
    (prod_path / "grouped.json").write_text(deterministic_dump(GroupedResultAdapter, grouped_analysis))
    (prod_path / "full.md").write_text(full_markdown)
    (prod_path / "summarized.md").write_text(summary_markdown)
    print(f"Wrote results to {prod_path}")



###############################################################################
# (c) Copyright 2024 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

from collections import defaultdict
from datetime import datetime, timedelta
from functools import cached_property

import pytz
from pydantic import BaseModel

from .constants import ACTIVE_TRANSFORMATION_STATES
from .gitlab import APGitlabRepo
from ..integrations import OperationsLogbook
from ..LogAnalysis.prod_analyzer import GroupedProblems


class APTransformDebugInfo(BaseModel):
    transform_id: int
    state: str
    type: str
    group_size: int
    active_inputs: set[int]
    file_counts: dict[str, int] | None = None
    problems: GroupedProblems | None = None


class APRequestDebugInfo(BaseModel):
    request_id: int
    transforms: list[APTransformDebugInfo] = []

    @cached_property
    def long_running(self):
        for transform in self.transforms:
            if transform.state not in ACTIVE_TRANSFORMATION_STATES:
                continue
            return bool(transform.active_inputs)
        return False


class APBaseDebugInfo(BaseModel):
    version: str
    issue_url: str | None
    requests: list[APRequestDebugInfo] = []


class APDebugInfo(APBaseDebugInfo):
    wg: str
    analysis: str
    created: datetime

    @cached_property
    def time_since_creation(self) -> timedelta:
        timezone = pytz.timezone("Europe/Zurich")
        return datetime.now(tz=timezone) - self.created

    @property
    def key(self):
        return (self.wg, self.analysis, self.version)


class APOverview(BaseModel):
    missing: list[int] = []
    recent: list[APDebugInfo] = []
    stuck: list[APDebugInfo] = []
    to_archive: list[APBaseDebugInfo] = []

    active: list[APDebugInfo] = []
    transform_ids_to_check: set[int] = set()

    @cached_property
    def running(self) -> list[APDebugInfo]:
        result = []
        for ap in self.active:
            requests = [req for req in ap.requests if not req.long_running]
            if not requests:
                continue
            ap = ap.model_copy(deep=True)
            ap.requests = requests
            result.append(ap)
        return result

    @cached_property
    def long_running(self) -> list[APDebugInfo]:
        result = []
        for ap in self.active:
            requests = [req for req in ap.requests if req.long_running]
            if not requests:
                continue
            ap = ap.model_copy(deep=True)
            ap.requests = requests
            result.append(ap)
        return result

    @cached_property
    def key_to_ap(self) -> dict[tuple[str, str, str], APDebugInfo]:
        result = {ap.key: ap for ap in self.active}
        assert len(self.active) == len(result)
        return result

    @cached_property
    def request_id_to_aps(self) -> dict[int, list[APDebugInfo]]:
        result = defaultdict(list)
        for ap in self.active:
            for req in ap.requests:
                result[req.request_id].append(ap)
        return dict(result)

    @cached_property
    def transform_id_to_request_id(self) -> dict[int, int]:
        result = {t.transform_id: req.request_id for ap in self.active for req in ap.requests for t in req.transforms}
        return result

    @cached_property
    def transform_id_to_transform(self) -> dict[int, APTransformDebugInfo]:
        result = {t.transform_id: t for ap in self.active for req in ap.requests for t in req.transforms}
        return result

    @cached_property
    def repo(self):
        return APGitlabRepo(with_auth=True)

    @cached_property
    def logbook(self):
        return OperationsLogbook()
